from typing import TYPE_CHECKING, List, Optional

from sqlalchemy.sql import text

from fittrackee import db
from fittrackee.files import get_absolute_file_path
from fittrackee.workouts.exceptions import WorkoutGPXException
from fittrackee.workouts.utils.geometry import (
    get_chart_data_from_segment_points,
)
from fittrackee.workouts.utils.gpx import get_chart_data_from_gpx

if TYPE_CHECKING:
    from fittrackee.workouts.models import Workout


def get_chart_data(
    workout: "Workout",
    *,
    can_see_heart_rate: bool,
    segment_id: Optional[int] = None,
) -> Optional[List]:
    """
    Get chart data from segments points if the segments have points, otherwise
    from the gpx file.
    """
    sql = """
        SELECT workout_segments.points
        FROM workout_segments
        WHERE workout_segments.workout_id  = :workout_id"""
    values = {"workout_id": workout.id}
    if segment_id is not None:
        segment_index = segment_id - 1
        if segment_index < 0:
            raise WorkoutGPXException("error", "Incorrect segment id", None)
        sql += """
          AND workout_segments.segment_id  = :segment_id"""
        values["segment_id"] = segment_index
    sql += """
        ORDER BY workout_id, segment_id"""
    segments_points = db.session.execute(text(sql), values).mappings().all()

    if not segments_points:
        raise WorkoutGPXException(
            "not found",
            (
                f"No segment with id '{segment_id}'"
                if segment_id is not None
                else "No segments"
            ),
            None,
        )

    if segments_points[0].points:
        return get_chart_data_from_segment_points(
            [segment["points"] for segment in segments_points],
            workout.sport.label,
            workout_ave_cadence=workout.ave_cadence,
            can_see_heart_rate=can_see_heart_rate,
        )

    if not workout.gpx:
        return []

    absolute_gpx_filepath = get_absolute_file_path(workout.gpx)
    return get_chart_data_from_gpx(
        absolute_gpx_filepath,
        workout.sport.label,
        workout.ave_cadence,
        can_see_heart_rate=can_see_heart_rate,
        segment_id=segment_id,
    )



from typing import Dict, List

from babel.support import Translations
from jinja2 import Environment, FileSystemLoader, select_autoescape


class I18nTemplate:
    def __init__(
        self,
        template_directory: str,
        translations_directory: str,
        languages: List[str],
    ) -> None:
        self._translations = self._get_translations(
            translations_directory, languages
        )
        self._env = Environment(
            autoescape=select_autoescape(["html", "htm", "xml"]),
            loader=FileSystemLoader(template_directory),
            extensions=["jinja2.ext.i18n"],
        )

    @staticmethod
    def _get_translations(
        translations_directory: str, languages: List[str]
    ) -> Dict:
        translations = {}
        for language in languages:
            translations[language] = Translations.load(
                dirname=translations_directory, locales=[language]
            )
        return translations

    def _load_translation(self, lang: str) -> None:
        self._env.install_gettext_translations(  # type: ignore
            self._translations[lang],
            newstyle=True,
        )

    def get_content(
        self, template_name: str, lang: str, part: str, data: Dict
    ) -> str:
        self._load_translation(lang)
        template = self._env.get_template(f"{template_name}/{part}")
        return template.render(data)

    def get_all_contents(
        self, template: str, lang: str, parts: List[str], data: Dict
    ) -> Dict:
        output = {}
        for part in parts:
            output[part] = self.get_content(template, lang, part, data)
        return output



from typing import Union

from flask import Blueprint, Response, request
from sqlalchemy import func

from fittrackee.responses import HttpResponse, UserNotFoundErrorResponse
from fittrackee.users.exceptions import UserNotFoundException
from fittrackee.users.models import User
from fittrackee.visibility_levels import VisibilityLevel
from fittrackee.workouts.models import Workout

from .feeds.workouts_feed_service import UserWorkoutsFeedService

feeds_blueprint = Blueprint("feeds", __name__)

FEED_ITEMS_LIMIT = 5


@feeds_blueprint.route(
    "/users/<string:user_name>/workouts.rss", methods=["GET"]
)
def get_user_public_workouts_rss_feed(
    user_name: str,
) -> Union[Response, HttpResponse]:
    """
    RSS feed with user's last 5 public workouts.

    Note: it does not display workouts when user is suspended.

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /user/Sam/workouts.rss HTTP/1.1

    - with some query parameters:

    .. sourcecode:: http

      GET /user/Sam/workouts.rss?lang=fr&imperial_units=true  HTTP/1.1

    :query string lang: RSS feed language (two-letter code, ISO 639-1,
           default: 'en'). If provided language is not supported, it falls
           back to English.
    :query boolean imperial_units: display values with imperial units.
           If false, metric system is used instead (default: false).
    :query boolean description: display workout description if true
           (default: false).

    :statuscode 200: ``success``
    :statuscode 404: ``user does not exist``

    """
    try:
        user = User.query.filter(
            func.lower(User.username) == func.lower(user_name),
        ).first()
        if not user:
            return UserNotFoundErrorResponse()
    except (ValueError, UserNotFoundException):
        return UserNotFoundErrorResponse()

    if user.suspended_at:
        latest_public_workouts = []
    else:
        latest_public_workouts = (
            Workout.query.filter(
                Workout.user_id == user.id,
                Workout.workout_visibility == VisibilityLevel.PUBLIC.value,
            )
            .order_by(Workout.workout_date.desc())
            .limit(FEED_ITEMS_LIMIT)
        ).all()

    params = request.args.copy()
    lang = params.get("lang", "en")
    use_imperial_units = (
        params.get("imperial_units", "false").lower() == "true"
    )
    with_description = params.get("description", "false").lower() == "true"

    feed_service = UserWorkoutsFeedService(
        user,
        latest_public_workouts,
        lang,
        use_imperial_units,
        with_description,
    )
    feed = feed_service.generate_user_workouts_feed()
    return Response(feed, mimetype="text/xml")



import os

from fittrackee.files import get_absolute_file_path


def get_upload_dir_size() -> int:
    """
    Return upload directory size
    """
    upload_path = get_absolute_file_path("")
    total_size = 0
    for dir_path, _, filenames in os.walk(upload_path):
        for f in filenames:
            fp = os.path.join(dir_path, f)
            total_size += os.path.getsize(fp)
    return total_size



import os
from typing import TYPE_CHECKING, Dict, Optional, Union

from fittrackee import appLog

from .visual_crossing import VisualCrossing

if TYPE_CHECKING:
    from ..workout_from_file.workout_point import WorkoutPoint


class WeatherService:
    """
    Available API:
    - VisualCrossing
    """

    def __init__(self) -> None:
        self.weather_api = self._get_weather_api()

    @staticmethod
    def _get_weather_api() -> Union["VisualCrossing", None]:
        weather_api_key: str = os.getenv("WEATHER_API_KEY", "")
        weather_api_provider: str = os.getenv(
            "WEATHER_API_PROVIDER", ""
        ).lower()

        if not weather_api_key:
            return None
        if weather_api_provider == "visualcrossing":
            return VisualCrossing(weather_api_key)
        return None

    def get_weather(self, point: "WorkoutPoint") -> Optional[Dict]:
        if not self.weather_api:
            return None
        try:
            return self.weather_api.get_weather(point)
        except Exception as e:
            appLog.error(f"error when getting weather data: {e}")
            return None



"""add open water swimming

Revision ID: 24eb097614e4
Revises: d22670a89a54
Create Date: 2023-07-19 10:34:40.297525

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '24eb097614e4'
down_revision = 'd22670a89a54'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Open Water Swimming', True, 0.1)
        """
    )

def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Open Water Swimming';
        """
    )



from datetime import timedelta
from typing import TYPE_CHECKING, Dict, List, TypedDict, Union

from fittrackee.equipments.utils import (
    SPORT_EQUIPMENT_TYPES,
    handle_equipments,
)
from fittrackee.visibility_levels import (
    VisibilityLevel,
    get_calculated_visibility,
)
from fittrackee.workouts.models import (
    DESCRIPTION_MAX_CHARACTERS,
    NOTES_MAX_CHARACTERS,
    TITLE_MAX_CHARACTERS,
    Sport,
)

from ..constants import WORKOUT_DATE_FORMAT
from ..exceptions import WorkoutException
from ..utils.workouts import get_workout_datetime
from .mixins import CheckWorkoutMixin

if TYPE_CHECKING:
    from fittrackee.equipments.models import Equipment
    from fittrackee.users.models import User

    from ..models import Workout


class WorkoutUpdateData(TypedDict, total=False):
    analysis_visibility: VisibilityLevel
    ascent: float
    descent: float
    description: str
    distance: float
    duration: int
    equipment_ids: List[str]
    map_visibility: VisibilityLevel
    notes: str
    sport_id: int
    title: str
    workout_date: str
    workout_visibility: VisibilityLevel


WITH_FILE_KEYS = {"analysis_visibility", "map_visibility"}
WITHOUT_FILE_KEYS = {
    "ascent",
    "descent",
    "distance",
    "duration",
    "workout_date",
}


class WorkoutUpdateService(CheckWorkoutMixin):
    def __init__(
        self, auth_user: "User", workout: "Workout", workout_data: Dict
    ):
        self.with_file = workout.gpx is not None
        self.workout_data = WorkoutUpdateData(**workout_data)
        self._check_workout_data()
        self.sport = self._get_sport()
        self.auth_user = auth_user
        self.workout = workout
        self.equipments_list = self._get_equipments_list()

    def _check_elevation_data(self) -> None:
        if self.with_file:
            return

        # for workout without gpx file, both elevation values must be
        # provided.
        try:
            if (
                (
                    "ascent" in self.workout_data
                    and "descent" not in self.workout_data
                )
                or (
                    "ascent" not in self.workout_data
                    and "descent" in self.workout_data
                )
            ) or (
                not (
                    self.workout_data.get("ascent") is None
                    and self.workout_data.get("descent") is None
                )
                and (
                    float(self.workout_data.get("ascent")) < 0  # type: ignore
                    or float(self.workout_data.get("descent")) < 0  # type: ignore
                )
            ):
                raise WorkoutException(
                    status="invalid",
                    message="invalid ascent or descent",
                )
        except (TypeError, ValueError) as e:
            raise WorkoutException(
                status="invalid",
                message="invalid ascent or descent",
            ) from e

    def _check_workout_data(self) -> None:
        invalid_keys = set(self.workout_data.keys()).intersection(
            WITHOUT_FILE_KEYS if self.with_file else WITH_FILE_KEYS
        )
        if invalid_keys:
            invalid_keys_str = ", ".join(
                sorted([f"'{key}'" for key in invalid_keys])
            )
            raise WorkoutException(
                status="invalid",
                message=(
                    f"invalid key{'' if len(invalid_keys) == 1 else 's'} "
                    f"({invalid_keys_str}) for workout with"
                    f"{'' if self.with_file else 'out'} gpx"
                ),
            )

        self._check_elevation_data()

    def _get_sport(self) -> Union["Sport", None]:
        sport_id = self.workout_data.get("sport_id")
        if sport_id:
            try:
                sport = Sport.query.filter_by(id=sport_id).one()
            except Exception as e:
                raise WorkoutException(
                    "invalid",
                    f"sport id {sport_id} not found",
                ) from e
            return sport
        return None

    def _get_equipments_list(self) -> Union[List["Equipment"], None]:
        if "equipment_ids" in self.workout_data:
            sport_id = (
                self.workout_data["sport_id"]
                if self.workout_data.get("sport_id")
                else self.workout.sport_id
            )
            return handle_equipments(
                self.workout_data["equipment_ids"],
                self.auth_user,
                sport_id,
                self.workout.equipments,
            )
        elif self.sport:
            # remove equipment if invalid for new sport
            # Note: for now only one equipment can be added
            for equipment in self.workout.equipments:
                if self.sport.label not in SPORT_EQUIPMENT_TYPES.get(
                    equipment.equipment_type.label, []
                ):
                    return []

        return None

    def _update_workout_without_file(self) -> None:
        if "ascent" in self.workout_data:
            self.workout.ascent = self.workout_data["ascent"]

        if "descent" in self.workout_data:
            self.workout.descent = self.workout_data["descent"]

        if self.workout_data.get("workout_date"):
            self.workout.workout_date, _ = get_workout_datetime(
                workout_date=self.workout_data["workout_date"],
                date_str_format=WORKOUT_DATE_FORMAT,
                user_timezone=self.auth_user.timezone,
            )

        update_speeds = False
        if self.workout_data.get("duration"):
            self.workout.duration = timedelta(
                seconds=self.workout_data["duration"]
            )
            self.workout.moving = self.workout.duration
            update_speeds = True

        if self.workout_data.get("distance"):
            self.workout.distance = self.workout_data["distance"]
            update_speeds = True

        if (
            update_speeds
            and self.workout.distance is not None
            and self.workout.duration is not None
        ):
            self.workout.ave_speed = (
                None
                if self.workout.duration.seconds == 0
                else float(self.workout.distance)
                / (self.workout.duration.seconds / 3600)
            )
            self.workout.max_speed = self.workout.ave_speed

        self._check_workout(self.workout)

    def update(self) -> None:
        if "sport_id" in self.workout_data:
            self.workout.sport_id = self.workout_data["sport_id"]
        if self.equipments_list is not None:
            self.workout.equipments = self.equipments_list

        if self.workout_data.get("description") is not None:
            self.workout.description = self.workout_data["description"][
                :DESCRIPTION_MAX_CHARACTERS
            ]
        if self.workout_data.get("notes") is not None:
            self.workout.notes = self.workout_data["notes"][
                :NOTES_MAX_CHARACTERS
            ]
        if self.workout_data.get("title") is not None:
            self.workout.title = self.workout_data["title"][
                :TITLE_MAX_CHARACTERS
            ]

        if "workout_visibility" in self.workout_data:
            self.workout.workout_visibility = self.workout_data[
                "workout_visibility"
            ]

        if self.with_file:
            if self.workout_data.get("analysis_visibility"):
                analysis_visibility = VisibilityLevel(
                    self.workout_data.get("analysis_visibility")
                )
                self.workout.analysis_visibility = get_calculated_visibility(
                    visibility=analysis_visibility,
                    parent_visibility=self.workout.workout_visibility,
                )
            if self.workout_data.get("map_visibility") is not None:
                map_visibility = VisibilityLevel(
                    self.workout_data.get("map_visibility")
                )
                self.workout.map_visibility = get_calculated_visibility(
                    visibility=map_visibility,
                    parent_visibility=self.workout.analysis_visibility,
                )
            return

        self._update_workout_without_file()



"""add standup paddleboarding

Revision ID: 684e59432ca4
Revises: 1648fe51cbd1
Create Date: 2025-06-15 14:38:00.820841

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '684e59432ca4'
down_revision = '1648fe51cbd1'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Standup Paddleboarding', True, 1);
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Standup Paddleboarding';
        """
    )



"""create Activities Segments table

Revision ID: dd73d23a7a3d
Revises: caf0e0dc621a
Create Date: 2018-05-14 12:12:57.299200

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'dd73d23a7a3d'
down_revision = 'caf0e0dc621a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('activity_segments',
    sa.Column('activity_id', sa.Integer(), nullable=False),
    sa.Column('segment_id', sa.Integer(), nullable=False),
    sa.Column('duration', sa.Interval(), nullable=False),
    sa.Column('pauses', sa.Interval(), nullable=True),
    sa.Column('moving', sa.Interval(), nullable=True),
    sa.Column('distance', sa.Numeric(precision=6, scale=3), nullable=True),
    sa.Column('min_alt', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.Column('max_alt', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.Column('descent', sa.Numeric(precision=7, scale=2), nullable=True),
    sa.Column('ascent', sa.Numeric(precision=7, scale=2), nullable=True),
    sa.Column('max_speed', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.Column('ave_speed', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.ForeignKeyConstraint(['activity_id'], ['activities.id'], ),
    sa.PrimaryKeyConstraint('activity_id', 'segment_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('activity_segments')
    # ### end Alembic commands ###



from typing import IO, Optional

import fitdecode
import gpxpy.gpx

from ...exceptions import WorkoutFileException
from .constants import GARMIN_DEVICES, NSMAP
from .workout_gpx_service import WorkoutGpxService


class WorkoutFitService(WorkoutGpxService):
    @staticmethod
    def get_coordinate(value: int) -> float:
        """
        converts coordinates from semicircles
        """
        return value * (180.0 / 2**31)

    @classmethod
    def parse_file(
        cls,
        workout_file: IO[bytes],
        segments_creation_event: str,
    ) -> "gpxpy.gpx.GPX":
        """
        For now only Activity Files are supported.
        see:
        https://developer.garmin.com/fit/file-types/activity/

        Activity File contains Laps (intervals in session) and Records.

        For now, only records are parsed and gpx file generated from fit file
        contains only one track. A new segment is created on after 'stop_all'
        event.

        TODO:
        - handle multiple sports activities (see Session)
        """
        try:
            fit_file = fitdecode.FitReader(workout_file)
        except Exception as e:
            raise WorkoutFileException(
                "error", "error when parsing fit file"
            ) from e

        gpx_track = gpxpy.gpx.GPXTrack()
        gpx_segment = gpxpy.gpx.GPXTrackSegment()
        creator: Optional[str] = None
        try:
            data_frames = filter(
                lambda frame: frame.frame_type == fitdecode.FIT_FRAME_DATA,
                fit_file,
            )

            # Handle device metadata from file_id
            file_id_frames = filter(
                lambda frame: frame.name == "file_id", data_frames
            )
            frame = next(file_id_frames, None)
            if frame and frame.has_field("manufacturer"):
                creator = frame.get_value("manufacturer")
                if frame.has_field("product") and frame.get_value("product"):
                    product = frame.get_raw_value("product")
                    if (
                        creator.lower() == "garmin"
                        and product in GARMIN_DEVICES.keys()
                    ):
                        product = GARMIN_DEVICES[product]
                    creator = f"{creator} {product}"

            # Handle the actual data frames. We sort them by timestamp
            # to handle devices that list events and records separately.
            event_and_record_frames = sorted(
                filter(
                    lambda frame: frame.name in ["event", "record"],
                    data_frames,
                ),
                key=lambda f: (
                    f.get_value("timestamp")
                    if f.has_field("timestamp")
                    else -1
                ),
            )

            for frame in event_and_record_frames:
                # create a new segment after 'stop_all' event
                if (
                    segments_creation_event in ["only_manual", "all"]
                    and frame.name == "event"
                    and frame.get_value("event") == "timer"
                    and frame.get_value("event_type") == "stop_all"
                ):
                    if (
                        segments_creation_event == "only_manual"
                        and frame.get_value("timer_trigger") != "manual"
                    ):
                        continue
                    if gpx_segment.points:
                        gpx_track.segments.append(gpx_segment)
                    gpx_segment = gpxpy.gpx.GPXTrackSegment()
                    continue

                if frame.name != "record":
                    continue

                longitude = (
                    frame.get_value("position_long")
                    if frame.has_field("position_long")
                    else None
                )
                latitude = (
                    frame.get_value("position_lat")
                    if frame.has_field("position_lat")
                    else None
                )
                time = (
                    frame.get_value("timestamp")
                    if frame.has_field("timestamp")
                    else None
                )
                if not longitude or not latitude or not time:
                    continue

                elevation = (
                    frame.get_value("enhanced_altitude")
                    if frame.has_field("enhanced_altitude")
                    else None
                )
                heart_rate = (
                    frame.get_value("heart_rate")
                    if frame.has_field("heart_rate")
                    else None
                )
                cadence = (
                    frame.get_value("cadence")
                    if frame.has_field("cadence")
                    else None
                )
                power = (
                    frame.get_value("power")
                    if frame.has_field("power")
                    else None
                )

                point = gpxpy.gpx.GPXTrackPoint(
                    longitude=cls.get_coordinate(longitude),
                    latitude=cls.get_coordinate(latitude),
                    elevation=float(elevation) if elevation else None,
                    time=time,
                )

                if any(
                    value is not None for value in [heart_rate, cadence, power]
                ):
                    point.extensions.append(
                        cls._get_extensions(heart_rate, cadence, power)
                    )
                gpx_segment.points.append(point)

        except fitdecode.exceptions.FitHeaderError as e:
            raise WorkoutFileException(
                "error", "error when parsing fit file"
            ) from e

        if gpx_segment.points:
            gpx_track.segments.append(gpx_segment)

        if not gpx_track.segments:
            raise WorkoutFileException(
                "error", "no valid segments with GPS found in fit file"
            ) from None

        gpx = gpxpy.gpx.GPX()
        gpx.creator = creator
        gpx.nsmap = NSMAP
        gpx.tracks.append(gpx_track)
        return gpx



# source for StandaloneApplication class:
# http://docs.gunicorn.org/en/stable/custom.html
import os
from typing import TYPE_CHECKING, Dict, Optional

import gunicorn.app.base

from fittrackee import create_app, db

if TYPE_CHECKING:
    from flask import Flask
    from gunicorn.arbiter import Arbiter
    from gunicorn.workers.base import Worker

HOST = os.getenv("HOST", "127.0.0.1")
PORT = os.getenv("PORT", "5000")
WORKERS = os.getenv("APP_WORKERS", 1)
TIMEOUT = os.getenv("APP_TIMEOUT", 30)
BASEDIR = os.path.abspath(os.path.dirname(__file__))
app = create_app()


def pre_fork(server: "Arbiter", worker: "Worker") -> None:
    """
    To use connection pools with a forked process (the new process will start
    with new connections).
    see: https://docs.sqlalchemy.org/en/20/core/pooling.html#using-connection-pools-with-multiprocessing-or-os-fork
    """
    with app.app_context():
        db.engine.dispose()


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    def __init__(
        self, current_app: "Flask", options: Optional[Dict] = None
    ) -> None:
        self.options = options or {}
        self.application = current_app
        super().__init__()

    def load_config(self) -> None:
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self) -> "Flask":
        return self.application


def main() -> None:
    options = {
        "bind": f"{HOST}:{PORT}",
        "workers": WORKERS,
        "pre_fork": pre_fork,
        "timeout": TIMEOUT,
    }
    StandaloneApplication(app, options).run()


if __name__ == "__main__":
    app.run()



"""update elevation precision

Revision ID: 0f375c44e659
Revises: a8cc0adfe1d3
Create Date: 2022-12-14 18:01:54.662987

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0f375c44e659'
down_revision = 'a8cc0adfe1d3'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        'workouts',
        'descent',
        existing_type=sa.NUMERIC(precision=7, scale=2),
        type_=sa.Numeric(precision=8, scale=3),
        existing_nullable=True,
    )
    op.alter_column(
        'workouts',
        'ascent',
        existing_type=sa.NUMERIC(precision=7, scale=3),
        type_=sa.Numeric(precision=8, scale=3),
        existing_nullable=True,
    )
    op.alter_column(
        'workout_segments',
        'descent',
        existing_type=sa.NUMERIC(precision=7, scale=2),
        type_=sa.Numeric(precision=8, scale=3),
        existing_nullable=True,
    )
    op.alter_column(
        'workout_segments',
        'ascent',
        existing_type=sa.NUMERIC(precision=7, scale=3),
        type_=sa.Numeric(precision=8, scale=3),
        existing_nullable=True,
    )


def downgrade():
    op.alter_column(
        'workout_segments',
        'ascent',
        existing_type=sa.NUMERIC(precision=8, scale=3),
        type_=sa.Numeric(precision=7, scale=2),
        existing_nullable=True,
    )
    op.alter_column(
        'workout_segments',
        'descent',
        existing_type=sa.NUMERIC(precision=8, scale=3),
        type_=sa.Numeric(precision=7, scale=2),
        existing_nullable=True,
    )
    op.alter_column(
        'workouts',
        'ascent',
        existing_type=sa.NUMERIC(precision=8, scale=3),
        type_=sa.Numeric(precision=7, scale=2),
        existing_nullable=True,
    )
    op.alter_column(
        'workouts',
        'descent',
        existing_type=sa.NUMERIC(precision=8, scale=3),
        type_=sa.Numeric(precision=7, scale=2),
        existing_nullable=True,
    )



from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from statistics import mean
from typing import IO, TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import gpxpy.gpx
import pytz
from lxml import etree as ET

from fittrackee import db

from ...exceptions import WorkoutExceedingValueException, WorkoutFileException
from ...models import WORKOUT_VALUES_LIMIT, Workout, WorkoutSegment
from .base_workout_with_segment_service import (
    BaseWorkoutWithSegmentsCreationService,
)
from .workout_point import WorkoutPoint

if TYPE_CHECKING:
    from uuid import UUID

    from fittrackee.users.models import User


@dataclass
class GpxInfo:
    distance: float
    duration: Optional[float]
    max_speed: float
    moving_time: float
    stopped_time: float
    max_alt: Optional[float]
    min_alt: Optional[float]
    ascent: Optional[float] = None
    descent: Optional[float] = None


def remove_microseconds(delta: "timedelta") -> "timedelta":
    return delta - timedelta(microseconds=delta.microseconds)


class WorkoutGpxService(BaseWorkoutWithSegmentsCreationService):
    def __init__(
        self,
        auth_user: "User",
        workout_file: IO[bytes],
        sport_id: int,
        stopped_speed_threshold: float,
        get_weather: bool = True,
        workout: Optional["Workout"] = None,
    ):
        super().__init__(
            auth_user,
            workout_file,
            sport_id,
            stopped_speed_threshold,
            workout,
            get_weather,
        )
        self.gpx: "gpxpy.gpx.GPX" = self.parse_file(
            workout_file, auth_user.segments_creation_event
        )
        self.cadences: List[int] = []
        self.heart_rates: List[int] = []
        self.powers: List[int] = []

    @staticmethod
    def _get_extensions(
        heart_rate: Optional[int], cadence: Optional[int], power: Optional[int]
    ) -> "ET.Element":
        track_point_extension = ET.Element("{gpxtpx}TrackPointExtension")
        if heart_rate is not None:
            heart_rate_element = ET.SubElement(
                track_point_extension, "{gpxtpx}hr"
            )
            heart_rate_element.text = str(heart_rate)
        if cadence is not None:
            cadence_element = ET.SubElement(
                track_point_extension, "{gpxtpx}cad"
            )
            cadence_element.text = str(cadence)
        if power is not None:
            power_element = ET.SubElement(
                track_point_extension, "{gpxtpx}power"
            )
            power_element.text = str(power)
        return track_point_extension

    @classmethod
    def parse_file(
        cls,
        workout_file: IO[bytes],
        segments_creation_event: str,
    ) -> "gpxpy.gpx.GPX":
        """
        Note:
        - segments_creation_event is not used (only for .fit files)
        """
        try:
            gpx = gpxpy.parse(workout_file)  # type: ignore
        except Exception as e:
            raise WorkoutFileException(
                "error", "error when parsing gpx file"
            ) from e
        if len(gpx.tracks) == 0:
            raise WorkoutFileException(
                "error", "no tracks in gpx file"
            ) from None
        return gpx

    @staticmethod
    def get_gpx_info(
        *,
        parsed_gpx: Union["gpxpy.gpx.GPXTrack", "gpxpy.gpx.GPXTrackSegment"],
        stopped_speed_threshold: float,
        use_raw_gpx_speed: bool,
    ) -> GpxInfo:
        moving_data = parsed_gpx.get_moving_data(
            stopped_speed_threshold=stopped_speed_threshold,
            raw=use_raw_gpx_speed,
        )
        if not moving_data:
            raise WorkoutFileException(
                "error", "gpx file is invalid"
            ) from None
        elevation = parsed_gpx.get_elevation_extremes()
        duration = parsed_gpx.get_duration()
        gpx_info = GpxInfo(
            duration=duration,
            distance=(
                moving_data.moving_distance + moving_data.stopped_distance
            ),
            moving_time=moving_data.moving_time,
            stopped_time=duration - moving_data.moving_time if duration else 0,
            max_speed=moving_data.max_speed,
            max_alt=elevation.maximum,
            min_alt=elevation.minimum,
        )
        if elevation.maximum:
            hill = parsed_gpx.get_uphill_downhill()
            gpx_info.ascent = hill.uphill
            gpx_info.descent = hill.downhill
        return gpx_info

    @staticmethod
    def check_gpx_info(gpx_info: "GpxInfo") -> None:
        for key, value in WORKOUT_VALUES_LIMIT.items():
            gpx_info_value = getattr(gpx_info, key)
            if gpx_info_value and gpx_info_value > value:
                raise WorkoutExceedingValueException(
                    f"'{key}' exceeds max value ({value})"
                )

    def set_calculated_data(
        self,
        *,
        parsed_gpx: Union["gpxpy.gpx.GPXTrack", "gpxpy.gpx.GPXTrackSegment"],
        object_to_update: Union["Workout", "WorkoutSegment"],
        stopped_time_between_segments: timedelta,
        stopped_speed_threshold: float,
        use_raw_gpx_speed: bool,
        hr_cadence_power_stats: dict,
    ) -> Union["Workout", "WorkoutSegment"]:
        gpx_info = self.get_gpx_info(
            parsed_gpx=parsed_gpx,
            stopped_speed_threshold=stopped_speed_threshold,
            use_raw_gpx_speed=use_raw_gpx_speed,
        )
        self.check_gpx_info(gpx_info)

        if isinstance(object_to_update, WorkoutSegment):
            object_to_update.max_speed = (gpx_info.max_speed / 1000) * 3600

        object_to_update.ascent = gpx_info.ascent
        object_to_update.ave_speed = (
            (
                gpx_info.distance / gpx_info.moving_time
                if gpx_info.moving_time > 0
                else 0
            )
            / 1000
            * 3600
        )
        object_to_update.descent = gpx_info.descent
        object_to_update.distance = gpx_info.distance / 1000
        object_to_update.duration = remove_microseconds(
            timedelta(seconds=gpx_info.duration if gpx_info.duration else 0)
            + stopped_time_between_segments
        )
        object_to_update.max_alt = gpx_info.max_alt
        object_to_update.min_alt = gpx_info.min_alt
        object_to_update.moving = remove_microseconds(
            timedelta(seconds=gpx_info.moving_time)
        )
        object_to_update.pauses = remove_microseconds(
            timedelta(seconds=gpx_info.stopped_time)
            + stopped_time_between_segments
        )

        object_to_update.ave_cadence = hr_cadence_power_stats["ave_cadence"]
        object_to_update.ave_hr = hr_cadence_power_stats["ave_hr"]
        object_to_update.ave_power = hr_cadence_power_stats["ave_power"]
        object_to_update.max_cadence = hr_cadence_power_stats["max_cadence"]
        object_to_update.max_hr = hr_cadence_power_stats["max_hr"]
        object_to_update.max_power = hr_cadence_power_stats["max_power"]

        return object_to_update

    def get_workout_date(self) -> "datetime":
        if not self.start_point or not self.start_point.time:
            raise WorkoutFileException(
                "error", "<time> is missing in gpx file"
            )
        return self.start_point.time.astimezone(timezone.utc)

    @staticmethod
    def _get_hr_cadence_power_data(
        heart_rates: List[int], cadences: List[int], powers: List[int]
    ) -> Dict:
        """
        Some files contains only zero cadence values. In this case, workout
        average and max cadences is None and cadence is not displayed.
        """
        ave_cadence = mean(cadences) if cadences else None
        return {
            "ave_cadence": ave_cadence if ave_cadence else None,
            "ave_hr": mean(heart_rates) if heart_rates else None,
            "ave_power": mean(powers) if powers else None,
            "max_cadence": max(cadences) if ave_cadence else None,
            "max_hr": max(heart_rates) if heart_rates else None,
            "max_power": max(powers) if powers else None,
        }

    def _process_segment_points(
        self,
        track_segment: "gpxpy.gpx.GPXTrackSegment",
        stopped_time_between_segments: timedelta,
        previous_segment_last_point_time: Optional[datetime],
        is_last_segment: bool,
        new_workout_segment: "WorkoutSegment",
        first_point: "gpxpy.gpx.GPXTrackPoint",
    ) -> Tuple[timedelta, Optional[datetime], Dict]:
        points = track_segment.points
        last_point_index = len(points) - 1
        cadences = []
        heart_rates = []
        powers = []
        previous_point = None
        previous_distance = 0.0
        segment_points: List[Dict] = []
        coordinates = []

        for point_idx, point in enumerate(points):
            if point_idx == 0:
                if not point.time:
                    raise WorkoutFileException(
                        "error", "<time> is missing in segment"
                    )
                new_workout_segment.start_date = point.time
                # if a previous segment exists, calculate stopped time
                # between the two segments
                if previous_segment_last_point_time and point.time:
                    stopped_time_between_segments += (
                        point.time - previous_segment_last_point_time
                    )

            distance = (
                point.distance_3d(previous_point)  # type: ignore[arg-type]
                if (
                    point.elevation
                    and previous_point
                    and previous_point.elevation
                )
                else point.distance_2d(previous_point)  # type: ignore[arg-type]
            )
            distance = 0.0 if distance is None else distance
            distance += previous_distance

            calculated_speed = track_segment.get_speed(point_idx)
            speed = (
                0.0
                if calculated_speed is None
                else round((calculated_speed / 1000) * 3600, 2)
            )

            time_difference = point.time_difference(first_point)
            segment_point: Dict = {
                "distance": distance,
                "duration": int(time_difference) if time_difference else 0,
                "elevation": point.elevation,
                "latitude": point.latitude,
                "longitude": point.longitude,
                "speed": speed,
                "time": (
                    str(point.time.astimezone(pytz.utc))
                    if point.time
                    else None
                ),
            }

            if point.extensions:
                extensions = []
                for extension in point.extensions:
                    if "TrackPointExtension" in extension.tag:
                        extensions.extend(extension)
                    else:
                        extensions.append(extension)
                for extension in extensions:
                    if not extension.text:
                        continue
                    if extension.tag == "power":
                        power = int(extension.text)
                        powers.append(power)
                        segment_point["power"] = power
                    if extension.tag.endswith("}hr"):
                        hr = int(extension.text)
                        heart_rates.append(hr)
                        segment_point["heart_rate"] = hr
                    if extension.tag.endswith("}cad"):
                        cadence = int(float(extension.text))
                        cadences.append(cadence)
                        segment_point["cadence"] = cadence
                    if extension.tag.endswith("}power"):
                        power = int(extension.text)
                        powers.append(power)
                        segment_point["power"] = power

            # last segment point
            if point_idx == last_point_index:
                previous_segment_last_point_time = point.time

                # last gpx point
                if is_last_segment and point.time:
                    self.end_point = WorkoutPoint(
                        point.longitude,
                        point.latitude,
                        point.time,
                    )
            coordinates.append([point.longitude, point.latitude])
            segment_points.append(segment_point)

            previous_point = point
            previous_distance = distance

        hr_cadence_stats = self._get_hr_cadence_power_data(
            heart_rates, cadences, powers
        )
        self.cadences.extend(cadences)
        self.heart_rates.extend(heart_rates)
        self.powers.extend(powers)
        self.coordinates.extend(coordinates)
        new_workout_segment.points = segment_points
        new_workout_segment.store_geometry(coordinates)

        return (
            stopped_time_between_segments,
            previous_segment_last_point_time,
            hr_cadence_stats,
        )

    def _process_segments(
        self,
        segments: List["gpxpy.gpx.GPXTrackSegment"],
        new_workout_id: int,
        new_workout_uuid: "UUID",
        first_point: "gpxpy.gpx.GPXTrackPoint",
    ) -> Tuple[timedelta, float]:
        last_segment_index = len(segments) - 1
        max_speed = 0.0
        previous_segment_last_point_time: Optional["datetime"] = None
        stopped_time_between_segments = timedelta(seconds=0)

        # remove existing segments if not creation
        if not self.is_creation and self.workout:
            WorkoutSegment.query.filter_by(workout_id=self.workout.id).delete()

        segment_idx = 0
        for segment in segments:
            # ignore segments with no distance
            if len(segment.points) < 2:
                continue

            new_workout_segment = WorkoutSegment(
                segment_id=segment_idx,
                workout_id=new_workout_id,
                workout_uuid=new_workout_uuid,
            )
            db.session.add(new_workout_segment)

            is_last_segment = segment_idx == last_segment_index
            (
                stopped_time_between_segments,
                previous_segment_last_point_time,
                hr_cadence_power_stats,
            ) = self._process_segment_points(
                segment,
                stopped_time_between_segments,
                previous_segment_last_point_time,
                is_last_segment,
                new_workout_segment,
                first_point,
            )

            self.set_calculated_data(
                parsed_gpx=segment,
                object_to_update=new_workout_segment,
                stopped_time_between_segments=timedelta(seconds=0),
                stopped_speed_threshold=self.stopped_speed_threshold,
                use_raw_gpx_speed=self.auth_user.use_raw_gpx_speed,
                hr_cadence_power_stats=hr_cadence_power_stats,
            )

            if (
                new_workout_segment.max_speed
                and new_workout_segment.max_speed > max_speed
            ):
                max_speed = new_workout_segment.max_speed

            segment_idx += 1

        return stopped_time_between_segments, max_speed

    def _process_file(self) -> "Workout":
        if not self.gpx:
            raise WorkoutFileException(
                "error", "no gpx, please load gpx file before"
            ) from None

        track: "gpxpy.gpx.GPXTrack" = self.gpx.tracks[0]

        start_point = None
        for segment in track.segments:
            # segment must contain at least 2 points to be valid.
            if len(segment.points) > 1:
                start_point = segment.points[0]
                break

        if not start_point:
            raise WorkoutFileException(
                "error", "no valid segments in file"
            ) from None

        if start_point.time:
            self.start_point = WorkoutPoint(
                start_point.longitude,
                start_point.latitude,
                start_point.time,
            )
        self.workout_name = track.name
        self.workout_description = track.description

        if not self.workout:
            self.workout = Workout(
                user_id=self.auth_user.id,
                sport_id=self.sport_id,
                workout_date=self.get_workout_date(),
            )
            db.session.add(self.workout)
            db.session.flush()
        self.workout.source = self.gpx.creator
        if self.start_point:
            self.workout.store_start_point_geometry(
                [self.start_point.longitude, self.start_point.latitude]
            )

        stopped_time_between_segments, max_speed = self._process_segments(
            track.segments, self.workout.id, self.workout.uuid, start_point
        )

        hr_cadence_power_stats = self._get_hr_cadence_power_data(
            self.heart_rates, self.cadences, self.powers
        )
        self.set_calculated_data(
            parsed_gpx=track,
            object_to_update=self.workout,
            stopped_time_between_segments=stopped_time_between_segments,
            stopped_speed_threshold=self.stopped_speed_threshold,
            use_raw_gpx_speed=self.auth_user.use_raw_gpx_speed,
            hr_cadence_power_stats=hr_cadence_power_stats,
        )
        self.workout.max_speed = max_speed
        bounds = track.get_bounds()
        self.workout.bounds = (
            [
                bounds.min_latitude,
                bounds.min_longitude,
                bounds.max_latitude,
                bounds.max_longitude,
            ]
            if (
                bounds
                and bounds.min_latitude
                and bounds.min_longitude
                and bounds.max_latitude
                and bounds.max_longitude
            )
            else []
        )

        return self.workout



import re
from datetime import timedelta
from typing import Optional, Union

from ..constants import POWER_SPORTS, RPM_CADENCE_SPORTS, SPM_CADENCE_SPORTS
from ..exceptions import InvalidDurationException


def convert_in_duration(value: str) -> timedelta:
    if not re.match(r"^([0-1]?\d|2[0-3]):[0-5]\d(:[0-5]\d)?$", value):
        raise InvalidDurationException()
    hours = int(value.split(":")[0])
    minutes = int(value.split(":")[1])
    return timedelta(seconds=(hours * 3600 + minutes * 60))


def convert_value_to_integer(
    record_type: str, val: Union[timedelta, float, None]
) -> Optional[int]:
    if val is None or record_type not in ["AS", "FD", "HA", "LD", "MS"]:
        return None

    if isinstance(val, timedelta):  # record_type == "LD"
        return int(val.total_seconds())

    multiplier = (
        100 if record_type in ["AS", "MS"] else 1000  # 'FD' and 'HA
    )
    return round(val * multiplier)


def get_cadence(sport_label: str, cadence: Optional[int]) -> Optional[int]:
    if cadence is None:
        return None

    if sport_label in RPM_CADENCE_SPORTS:
        return cadence

    if sport_label in SPM_CADENCE_SPORTS:
        return cadence * 2

    return None


def get_power(sport_label: str, power: Optional[int]) -> Optional[int]:
    if power is None:
        return None

    if sport_label in POWER_SPORTS:
        return power

    return None



from typing import Dict, Union

from flask import Blueprint, request
from sqlalchemy import func

from fittrackee import db
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import HttpResponse, InvalidPayloadErrorResponse
from fittrackee.users.models import TASK_TYPES, User, UserTask
from fittrackee.users.roles import UserRole

queued_tasks_blueprint = Blueprint("queued_users_tasks", __name__)

TASKS_PER_PAGE = 5
USERS_QUEUED_TASKS_FILTERS = [
    UserTask.progress == 0,
    UserTask.errored == False,  # noqa
    UserTask.aborted == False,  # noqa
]


@queued_tasks_blueprint.route("/users/tasks/queued", methods=["GET"])
@require_auth(scopes=["users:read"], role=UserRole.ADMIN)
def get_queued_users_tasks_count(
    auth_user: "User",
) -> Union[Dict, HttpResponse]:
    """
    Get queued tasks.

    **Scope**: ``tasks:read``

    **Minimum role**: Administrator

    **Example request**:

    .. sourcecode:: http

      GET /api/tasks/queued HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "counts": {
          "user_data_export": 0,
          "workouts_archive_upload": 1
        },
        "status": "success"
      }

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    """
    tasks_counts = (
        db.session.query(UserTask.task_type, func.count(UserTask.task_type))
        .filter(*USERS_QUEUED_TASKS_FILTERS)
        .group_by(UserTask.task_type)
        .all()
    )
    counts: Dict = {"user_data_export": 0, "workouts_archive_upload": 0}
    for count in tasks_counts:
        counts[count[0]] = count[1]

    return {
        "status": "success",
        "counts": counts,
    }


@queued_tasks_blueprint.route(
    "/users/tasks/queued/<string:task_type>", methods=["GET"]
)
@require_auth(scopes=["users:read"], role=UserRole.ADMIN)
def get_queued_users_tasks_list(
    auth_user: "User", task_type: str
) -> Union[Dict, HttpResponse]:
    """
    Get queued tasks for a given type

    **Scope**: ``tasks:read``

    **Minimum role**: Administrator

    **Example request**:

    .. sourcecode:: http

      GET /api/config HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "pagination": {
          "has_next": false,
          "has_prev": false,
          "page": 1,
          "pages": 1,
          "total": 1
        },
        "queued_tasks": [
          {
            "created_at": "Mon, 21 Apr 2025 14:24:53 GMT",
            "file_size": 21750,
            "files_count": 30,
            "id": "5RnDqwLuANmDKRqFhkgJsh",
            "message_id": "3542fd2d-e8f7-438f-b018-0fd49176728a",
            "status": "queued",
            "type": "workouts_archive_upload",
            "user": {
              "blocked": false,
              "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
              "email": "sam@example.com",
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_active": true,
              "is_followed_by": "false",
              "nb_workouts": 1,
              "picture": false,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            }
          }
        ],
        "status": "success"
      }

    :statuscode 200: ``success``
    :statuscode 400: ``invalid task type``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403: ``you do not have permissions``
    """
    if task_type not in TASK_TYPES:
        return InvalidPayloadErrorResponse("invalid task type", "invalid")

    params = request.args.copy()
    page = int(params.get("page", 1))

    queued_tasks_pagination = (
        db.session.query(UserTask, User)  # type: ignore
        .join(User, User.id == UserTask.user_id)
        .filter(*USERS_QUEUED_TASKS_FILTERS, UserTask.task_type == task_type)
        .order_by(UserTask.created_at.desc())
        .paginate(page=page, per_page=TASKS_PER_PAGE, error_out=False)
    )

    return {
        "queued_tasks": [
            task.serialize(
                current_user=auth_user, for_admin=True, task_user=task_user
            )
            for task, task_user in queued_tasks_pagination.items
        ],
        "status": "success",
        "pagination": {
            "has_next": queued_tasks_pagination.has_next,
            "has_prev": queued_tasks_pagination.has_prev,
            "page": queued_tasks_pagination.page,
            "pages": queued_tasks_pagination.pages,
            "total": queued_tasks_pagination.total,
        },
    }



from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import pytz

from fittrackee import db
from fittrackee.visibility_levels import VisibilityLevel

from ..constants import WORKOUT_DATE_FORMAT
from ..exceptions import WorkoutException
from ..models import (
    DESCRIPTION_MAX_CHARACTERS,
    NOTES_MAX_CHARACTERS,
    Workout,
)
from .base_workout_service import BaseWorkoutService
from .mixins import CheckWorkoutMixin

if TYPE_CHECKING:
    from fittrackee.users.models import User


@dataclass
class WorkoutData:
    distance: float
    duration: int
    sport_id: int
    workout_date: str
    ascent: Optional[float] = None
    descent: Optional[float] = None
    description: Optional[str] = None
    equipment_ids: Optional[List[str]] = None
    notes: Optional[str] = None
    title: Optional[str] = None
    workout_visibility: Optional[VisibilityLevel] = None


class WorkoutCreationService(CheckWorkoutMixin, BaseWorkoutService):
    """
    Workout creation without file
    """

    def __init__(self, auth_user: "User", workout_data: Dict):
        super().__init__(
            auth_user,
            workout_data["sport_id"],
            workout_data.get("equipment_ids"),
        )
        self.workout_data = WorkoutData(**workout_data)

    def get_workout_date(self) -> "datetime":
        try:
            workout_date = datetime.strptime(  # noqa: DTZ007
                self.workout_data.workout_date, WORKOUT_DATE_FORMAT
            )
        except ValueError as e:
            raise WorkoutException(
                "error", "invalid format for workout date"
            ) from e
        if self.auth_user.timezone:
            workout_date = pytz.timezone(self.auth_user.timezone).localize(
                workout_date
            )
            return workout_date.astimezone(pytz.utc)

        # in case no timezone is set for authenticated user
        return workout_date.replace(tzinfo=timezone.utc)

    def _get_elevation(self) -> Tuple[Optional[float], Optional[float]]:
        ascent = self.workout_data.ascent
        descent = self.workout_data.descent
        try:
            if (
                (ascent is None and descent is not None)
                or (ascent is not None and descent is None)
                or (ascent is not None and float(ascent) < 0)
                or (descent is not None and float(descent) < 0)
            ):
                raise WorkoutException("invalid", "invalid ascent or descent")
        except ValueError as e:
            raise WorkoutException(
                "invalid", "invalid ascent or descent"
            ) from e
        return ascent, descent

    def _get_workout_title(self, workout_date: datetime) -> str:
        return self._get_title(workout_date, self.workout_data.title)

    def process(self) -> Tuple[List["Workout"], Dict]:
        ascent, descent = self._get_elevation()
        duration = timedelta(seconds=self.workout_data.duration)
        distance = self.workout_data.distance

        workout_date = self.get_workout_date()
        new_workout = Workout(
            distance=distance,
            duration=duration,
            sport_id=self.workout_data.sport_id,
            user_id=self.auth_user.id,
            workout_date=workout_date,
        )
        db.session.add(new_workout)
        new_workout.moving = duration
        new_workout.pauses = timedelta(seconds=0)
        new_workout.ave_speed = (
            None
            if duration.seconds == 0
            else float(distance) / (duration.seconds / 3600)
        )
        new_workout.max_speed = new_workout.ave_speed
        new_workout.ascent = ascent
        new_workout.descent = descent

        self._check_workout(new_workout)
        new_workout.title = self._get_workout_title(workout_date)
        new_workout.description = (
            self.workout_data.description[:DESCRIPTION_MAX_CHARACTERS]
            if self.workout_data.description
            else None
        )
        new_workout.notes = (
            None
            if self.workout_data.notes is None
            else self.workout_data.notes[:NOTES_MAX_CHARACTERS]
        )

        equipments = self.get_equipments()
        if equipments is not None:
            new_workout.equipments = equipments

        new_workout.workout_visibility = (
            self.workout_data.workout_visibility
            if self.workout_data.workout_visibility
            else self.auth_user.workouts_visibility
        )

        db.session.flush()
        return [new_workout], {}



"""add static map id to 'Activity' table

Revision ID: 9f8c9c37da44
Revises: 5a42db64e872
Create Date: 2018-05-30 12:48:11.714627

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9f8c9c37da44'
down_revision = '5a42db64e872'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('activities', sa.Column('map_id', sa.String(length=50), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('activities', 'map_id')
    # ### end Alembic commands ###



import logging
import sys
from typing import Optional

import click
from humanize import naturalsize

from fittrackee import db
from fittrackee.cli.app import app
from fittrackee.languages import SUPPORTED_LANGUAGES
from fittrackee.users.exceptions import UserNotFoundException
from fittrackee.users.export_data import (
    clean_user_data_export,
    generate_user_data_archives,
    process_queued_data_export,
)
from fittrackee.users.roles import UserRole
from fittrackee.users.timezones import get_timezone
from fittrackee.users.users_service import UserManagerService
from fittrackee.users.utils.language import get_language
from fittrackee.users.utils.tokens import clean_blacklisted_tokens

handler = logging.StreamHandler()
logger = logging.getLogger("fittrackee_users_cli")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


@click.group(name="users")
def users_cli() -> None:
    """Manage users."""
    pass


@users_cli.command("create")
@click.argument("username")
@click.option("--email", type=str, required=True, help="User email.")
@click.option(
    "--password",
    type=str,
    help="User password. If not provided, a random password is generated.",
)
@click.option(
    "--lang",
    type=str,
    help=(
        "User preference for interface language (two-letter code, ISO 639-1)."
        " If not provided or not supported, it falls back to English ('en')."
        f" Supported languages: {', '.join(SUPPORTED_LANGUAGES)}."
    ),
)
@click.option(
    "--tz",
    type=str,
    help=(
        "User preference for timezone. If not provided or not supported, "
        "it falls back to 'Europe/Paris'."
    ),
)
@click.option(
    "--role", type=click.Choice(UserRole.db_choices()), help="Set user role."
)
def create_user(
    username: str,
    email: str,
    password: Optional[str],
    lang: Optional[str],
    role: Optional[str],
    tz: Optional[str],
) -> None:
    """Create an active user account."""
    with app.app_context():
        try:
            user_manager_service = UserManagerService(username)
            user, user_password = user_manager_service.create_user(
                email=email, password=password, check_email=True, role=role
            )
            if user:
                db.session.add(user)
                user_language = get_language(lang)
                user.language = user_language
                user_timezone = get_timezone(tz)
                user.timezone = user_timezone
                db.session.commit()
                user_manager_service.update(activate=True)
                click.echo(f"User '{username}' created.")
                if lang != user_language:
                    click.echo(
                        "The user preference for interface language "
                        f"is: {user_language}"
                    )
                if tz != user_timezone:
                    click.echo(
                        f"The user preference for timezone is: {user_timezone}"
                    )
                if not password:
                    click.echo(f"The user password is: {user_password}")
        except Exception as e:
            click.echo(f"Error(s) occurred:\n{e}", err=True)


@users_cli.command("update")
@click.argument("username")
@click.option(
    "--set-role",
    type=click.Choice(UserRole.db_choices()),
    help="Set user role (when setting 'moderator', 'admin' and 'owner' "
    "role, it also activates user account if not active).",
)
@click.option("--activate", is_flag=True, help="Activate user account.")
@click.option(
    "--reset-password",
    is_flag=True,
    help="Reset user password (a new password will be displayed).",
)
@click.option("--update-email", type=str, help="Update user email.")
def manage_user(
    username: str,
    set_role: Optional[str],
    activate: bool,
    reset_password: bool,
    update_email: Optional[str],
) -> None:
    """Manage given user account."""
    with app.app_context():
        try:
            user_manager_service = UserManagerService(username)
            _, is_user_updated, password, _ = user_manager_service.update(
                role=set_role,
                with_confirmation=False,
                activate=activate if activate else None,
                reset_password=reset_password,
                new_email=update_email,
            )
            if is_user_updated:
                click.echo(f"User '{username}' updated.")
                if password:
                    click.echo(f"The new password is: {password}")
            else:
                click.echo("No updates.")
        except UserNotFoundException:
            click.echo(
                f"User '{username}' not found.\n"
                "Check the provided user name (case sensitive).",
                err=True,
            )
        except Exception as e:
            click.echo(f"An error occurred: {e}", err=True)


@users_cli.command("clean_tokens")
@click.option("--days", type=int, required=True, help="Number of days.")
def clean(
    days: int,
) -> None:
    """
    Clean blacklisted tokens expired for more than provided number of days.
    """
    with app.app_context():
        deleted_rows = clean_blacklisted_tokens(days)
        logger.info(f"Blacklisted tokens deleted: {deleted_rows}.")


@users_cli.command("clean_archives")
@click.option("--days", type=int, required=True, help="Number of days.")
def clean_export_archives(
    days: int,
) -> None:
    """
    Clean user export archives created for more than provided number of days.
    """
    with app.app_context():
        counts = clean_user_data_export(days)
        logger.info(
            f"Deleted data export requests: {counts['deleted_requests']}."
        )
        logger.info(
            f"Deleted data export archives: {counts['deleted_archives']}."
        )
        logger.info(f"Freed space: {naturalsize(counts['freed_space'])}.")


@users_cli.command("export_archives")
@click.option(
    "--max",
    "max_reports",
    type=int,
    required=True,
    help="Maximum number of export requests to process.",
)
def export_archives(max_reports: int) -> None:
    """
    Export user data in zip archive if queued requests exist.

    To use in case redis is not set.
    """
    with app.app_context():
        count = generate_user_data_archives(max_reports)
        logger.info(f"Generated data export archives: {count}.")


@users_cli.command("export_archive")
@click.option(
    "--id",
    "task_id",
    type=str,
    required=True,
    help="Id of task to process",
)
def export_archive(task_id: str) -> None:
    """
    Process queued user data export.

    To use in case redis is not set.
    """
    with app.app_context():
        try:
            process_queued_data_export(task_id)
        except Exception as e:
            logger.error(str(e))
            sys.exit(1)
        logger.info("\nDone.")



from datetime import datetime, timezone
from typing import Dict, Optional, Tuple, Union

from flask import Blueprint, request
from sqlalchemy import exc

from fittrackee import db
from fittrackee.oauth2.server import require_auth
from fittrackee.reports.models import ReportActionAppeal
from fittrackee.responses import (
    ForbiddenErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.models import User
from fittrackee.utils import clean_input
from fittrackee.visibility_levels import VisibilityLevel
from fittrackee.workouts.decorators import check_workout
from fittrackee.workouts.models import Workout

from .decorators import check_workout_comment
from .models import Comment, CommentLike, get_comments

comments_blueprint = Blueprint("comments", __name__)

DEFAULT_COMMENT_LIKES_PER_PAGE = 10


@comments_blueprint.route(
    "/workouts/<string:workout_short_id>/comments", methods=["POST"]
)
@require_auth(scopes=["workouts:write"])
@check_workout(only_owner=False, as_data=False)
def post_workout_comment(
    auth_user: User, workout: Workout, workout_short_id: str
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Post a comment.

    **Scope**: ``workouts:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/workouts/2oRDfncv6vpRkfp3yrCYHt HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "comment": {
            "created_at": "Sun, 01 Dec 2024 13:45:34 GMT",
            "id": "WJgTwtqFpnPrHYAK5eX9Pw",
            "liked": false,
            "likes_count": 0,
            "mentions": [],
            "modification_date": null,
            "suspended_at": null,
            "text": "Great!",
            "text_html": "Great!",
            "text_visibility": "private",
            "user": {
              "created_at": "Sun, 24 Nov 2024 16:52:14 GMT",
              "followers": 3,
              "following": 2,
              "nb_workouts": 10,
              "picture": true,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "workout_id": "2oRDfncv6vpRkfp3yrCYHt"
          },
          "status": "created"
        }

    :param string workout_short_id: workout short id

    :<json string text: comment content
    :<json string text_visibility: visibility level (``public``,
           ``followers_only``, ``private``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 201: ``created``
    :statuscode 400:
        - ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``workout not found``
    :statuscode 500: ``Error during comment save.``
    """
    comment_data = request.get_json()
    if (
        not comment_data
        or not comment_data.get("text")
        or not comment_data.get("text_visibility")
    ):
        return InvalidPayloadErrorResponse()
    try:
        new_comment = Comment(
            user_id=auth_user.id,
            workout_id=workout.id,
            text=clean_input(comment_data["text"]),
            text_visibility=VisibilityLevel(comment_data["text_visibility"]),
        )
        db.session.add(new_comment)
        db.session.flush()
        new_comment.create_mentions()
        db.session.commit()

        return (
            {
                "status": "created",
                "comment": new_comment.serialize(auth_user),
            },
            201,
        )
    except (exc.IntegrityError, ValueError) as e:
        return handle_error_and_return_response(
            error=e,
            message="Error during comment save.",
            status="fail",
            db=db,
        )


@comments_blueprint.route(
    "/comments/<string:comment_short_id>", methods=["GET"]
)
@require_auth(scopes=["workouts:read"], optional_auth_user=True)
@check_workout_comment(only_owner=False)
def get_workout_comment(
    auth_user: Optional[User], comment: Comment
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Get comment.

    **Scope**: ``workouts:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/workouts/2oRDfncv6vpRkfp3yrCYHt/comment HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "comment": {
            "created_at": "Sun, 01 Dec 2024 13:45:34 GMT",
            "id": "T2zeeUXvuy3PLA8MeeUFyk",
            "liked": false,
            "likes_count": 0,
            "mentions": [],
            "modification_date": null,
            "suspended_at": null,
            "text": "Nice!",
            "text_html": "Nice!",
            "text_visibility": "private",
            "user": {
              "created_at": "Sun, 24 Nov 2024 16:52:14 GMT",
              "followers": 3,
              "following": 2,
              "nb_workouts": 10,
              "picture": true,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "workout_id": null
          },
          "status": "success"
        }

    :param string comment_short_id: comment short id

    :reqheader Authorization: OAuth 2.0 Bearer Token for comment with
        ``private`` and ``followers_only`` visibility

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``workout comment not found``
    """
    return (
        {
            "status": "success",
            "comment": comment.serialize(auth_user),
        },
        200,
    )


@comments_blueprint.route(
    "/workouts/<string:workout_short_id>/comments", methods=["GET"]
)
@require_auth(scopes=["workouts:read"], optional_auth_user=True)
@check_workout(only_owner=False, as_data=False)
def get_workout_comments(
    auth_user: Optional[User], workout: Workout, workout_short_id: str
) -> Union[Dict, HttpResponse]:
    """
    Get workout comments.

    It returns only comments visible to authenticated user and only public
    comments when no authentication provided.

    **Scope**: ``workouts:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/workouts/2oRDfncv6vpRkfp3yrCYHt/comments HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "comments": [
              {
                "created_at": "Sun, 01 Dec 2024 13:45:34 GMT",
                "id": "WJgTwtqFpnPrHYAK5eX9Pw",
                "liked": false,
                "likes_count": 0,
                "mentions": [],
                "modification_date": null,
                "suspended_at": null,
                "text": "Great!",
                "text_html": "Great!",
                "text_visibility": "private",
                "user": {
                  "created_at": "Sun, 24 Nov 2024 16:52:14 GMT",
                  "followers": 3,
                  "following": 2,
                  "nb_workouts": 10,
                  "picture": true,
                  "role": "user",
                  "suspended_at": null,
                  "username": "Sam"
                },
                "workout_id": "2oRDfncv6vpRkfp3yrCYHt"
              }
            ]
          },
          "status": "success"
        }

    :param string workout_short_id: workout short id

    :reqheader Authorization: OAuth 2.0 Bearer Token for comments with
        ``private`` and ``followers_only`` visibility

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``workout not found``
    :statuscode 500: ``Error during comment save.``
    """
    try:
        comments = get_comments(
            workout_id=workout.id,
            user=auth_user,
        )
        return {
            "status": "success",
            "data": {
                "comments": [
                    comment.serialize(auth_user) for comment in comments
                ]
            },
        }
    except Exception as e:
        return handle_error_and_return_response(e)


@comments_blueprint.route(
    "/comments/<string:comment_short_id>", methods=["DELETE"]
)
@require_auth(scopes=["workouts:write"])
@check_workout_comment()
def delete_workout_comment(
    auth_user: User, comment: Comment
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Delete workout comment.

    **Scope**: ``workouts:write``

    **Example request**:

    .. sourcecode:: http

      DELETE /api/comments/MzydiCYYfktG3gga2x8AfU HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 NO CONTENT
      Content-Type: application/json

    :param string comment_short_id: comment short id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: comment deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``comment not found``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    try:
        db.session.delete(comment)
        db.session.commit()
        return {"status": "no content"}, 204
    except (
        exc.IntegrityError,
        exc.OperationalError,
        ValueError,
        OSError,
    ) as e:
        return handle_error_and_return_response(e, db=db)


@comments_blueprint.route(
    "/comments/<string:comment_short_id>", methods=["PATCH"]
)
@require_auth(scopes=["workouts:write"])
@check_workout_comment()
def update_workout_comment(
    auth_user: User, comment: Comment
) -> Union[Dict, HttpResponse]:
    """
    Update comment text.

    **Scope**: ``workouts:write``

    **Example request**:

    .. sourcecode:: http

      PATCH /api/workouts/WJgTwtqFpnPrHYAK5eX9Pw HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "comment": {
            "created_at": "Sun, 01 Dec 2024 13:45:34 GMT",
            "id": "WJgTwtqFpnPrHYAK5eX9Pw",
            "liked": false,
            "likes_count": 0,
            "mentions": [],
            "modification_date": null,
            "suspended_at": null,
            "text": "Great!",
            "text_html": "Great!",
            "text_visibility": "private",
            "user": {
              "created_at": "Sun, 24 Nov 2024 16:52:14 GMT",
              "followers": 3,
              "following": 2,
              "nb_workouts": 10,
              "picture": true,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "workout_id": "2oRDfncv6vpRkfp3yrCYHt"
          },
          "status": "success"
        }

    :param string comment_short_id: comment short id

    :<json string text: comment content

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``comment not found``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    comment_data = request.get_json()
    if not comment_data or not comment_data.get("text"):
        return InvalidPayloadErrorResponse()

    try:
        comment.text = clean_input(comment_data["text"])
        comment.modification_date = datetime.now(timezone.utc)
        comment.update_mentions()
        db.session.commit()
        return {
            "status": "success",
            "comment": comment.serialize(auth_user),
        }

    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e)


@comments_blueprint.route(
    "/comments/<string:comment_short_id>/like", methods=["POST"]
)
@require_auth(scopes=["workouts:write"])
@check_workout_comment(only_owner=False)
def like_comment(
    auth_user: User, comment: Comment
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Add a "like" to a comment.

    **Scope**: ``workouts:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/comments/WJgTwtqFpnPrHYAK5eX9Pw/like HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "comment": {
            "created_at": "Sun, 01 Dec 2024 13:45:34 GMT",
            "id": "WJgTwtqFpnPrHYAK5eX9Pw",
            "liked": true,
            "likes_count": 1,
            "mentions": [],
            "modification_date": null,
            "suspended_at": null,
            "text": "Great!",
            "text_html": "Great!",
            "text_visibility": "private",
            "user": {
              "created_at": "Sun, 24 Nov 2024 16:52:14 GMT",
              "followers": 3,
              "following": 2,
              "nb_workouts": 10,
              "picture": true,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "workout_id": "2oRDfncv6vpRkfp3yrCYHt"
          },
          "status": "success"
        }

    :param string comment_short_id: comment short id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``comment not found``
    """
    if comment.suspended_at:
        return ForbiddenErrorResponse()
    try:
        like = CommentLike(user_id=auth_user.id, comment_id=comment.id)
        db.session.add(like)
        db.session.commit()
    except exc.IntegrityError:
        db.session.rollback()
    return {
        "status": "success",
        "comment": comment.serialize(auth_user),
    }, 200


@comments_blueprint.route(
    "/comments/<string:comment_short_id>/like/undo",
    methods=["POST"],
)
@require_auth(scopes=["workouts:write"])
@check_workout_comment(only_owner=False)
def undo_comment_like(
    auth_user: User, comment: Comment
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Remove a comment "like".

    **Scope**: ``workouts:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/comments/2oRDfncv6vpRkfp3yrCYHt/like/undo HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "comment": {
            "created_at": "Sun, 01 Dec 2024 13:45:34 GMT",
            "id": "WJgTwtqFpnPrHYAK5eX9Pw",
            "liked": false,
            "likes_count": 0,
            "mentions": [],
            "modification_date": null,
            "suspended_at": null,
            "text": "Great!",
            "text_html": "Great!",
            "text_visibility": "private",
            "user": {
              "created_at": "Sun, 24 Nov 2024 16:52:14 GMT",
              "followers": 3,
              "following": 2,
              "nb_workouts": 10,
              "picture": true,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "workout_id": "2oRDfncv6vpRkfp3yrCYHt"
          },
          "status": "success"
        }

    :param string comment_short_id: comment short id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``comment not found``
    """
    like = CommentLike.query.filter_by(
        user_id=auth_user.id, comment_id=comment.id
    ).first()
    if like:
        db.session.delete(like)
        db.session.commit()
    return {
        "status": "success",
        "comment": comment.serialize(auth_user),
    }, 200


@comments_blueprint.route(
    "/comments/<string:comment_short_id>/likes", methods=["GET"]
)
@require_auth(scopes=["workouts:read"], optional_auth_user=True)
@check_workout_comment(only_owner=False)
def get_comment_likes(
    auth_user: User, comment: Comment
) -> Union[Dict, HttpResponse]:
    """
    Get users who like comment.

    **Scope**: ``workouts:read``

    **Example request**:

    .. sourcecode:: http

      POST /api/comments/WJgTwtqFpnPrHYAK5eX9Pw/likes HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "data": {
            "likes": [
              {
                "created_at": "Sun, 31 Dec 2017 09:00:00 GMT",
                "followers": 0,
                "following": 0,
                "nb_workouts": 1,
                "picture": false,
                "role": "user",
                "suspended_at": null,
                "username": "Sam"
              }
            ]
          },
          "status": "success"
        }

    :param string comment_short_id: comment short id

    :query integer page: page if using pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token for comment with
        ``private`` and ``followers_only`` visibility

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``comment not found``
    """
    params = request.args.copy()
    page = int(params.get("page", 1))
    likes_pagination = (
        User.query.join(CommentLike, User.id == CommentLike.user_id)
        .filter(CommentLike.comment_id == comment.id)
        .order_by(CommentLike.created_at.desc())
        .paginate(
            page=page, per_page=DEFAULT_COMMENT_LIKES_PER_PAGE, error_out=False
        )
    )
    users = likes_pagination.items
    return {
        "status": "success",
        "data": {
            "likes": [user.serialize(current_user=auth_user) for user in users]
        },
        "pagination": {
            "has_next": likes_pagination.has_next,
            "has_prev": likes_pagination.has_prev,
            "page": likes_pagination.page,
            "pages": likes_pagination.pages,
            "total": likes_pagination.total,
        },
    }


@comments_blueprint.route(
    "/comments/<string:comment_short_id>/suspension/appeal",
    methods=["POST"],
)
@require_auth(scopes=["workouts:write"])
@check_workout_comment(only_owner=True)
def appeal_comment_suspension(
    auth_user: User, comment: Comment
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Appeal comment suspension.

    Only comment author can appeal the suspension.

    **Scope**: ``workouts:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/comments/WJgTwtqFpnPrHYAK5eX9Pw/suspension/appeal HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 201 CREATED
      Content-Type: application/json

        {
          "status": "success"
        }

    :param string comment_short_id: comment short id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 201: appeal created
    :statuscode 400:
        - ``no text provided``
        - ``you can appeal only once``
        - ``workout comment is not suspended``
        - ``workout comment has no suspension``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``comment not found``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    if not comment.suspended_at:
        return InvalidPayloadErrorResponse("workout comment is not suspended")
    suspension_action = comment.suspension_action
    if not suspension_action:
        return InvalidPayloadErrorResponse("workout comment has no suspension")

    text = request.get_json().get("text")
    if not text:
        return InvalidPayloadErrorResponse("no text provided")

    try:
        appeal = ReportActionAppeal(
            action_id=suspension_action.id, user_id=auth_user.id, text=text
        )
        db.session.add(appeal)
        db.session.commit()
        return {"status": "success"}, 201

    except exc.IntegrityError:
        return InvalidPayloadErrorResponse("you can appeal only once")
    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)



"""add snowshoes sport

Revision ID: ed409fd9db9d
Revises: 07188ca7620a
Create Date: 2021-12-19 09:09:37.531543

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'ed409fd9db9d'
down_revision = '07188ca7620a'
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        """
        INSERT INTO sports (label, is_active, stopped_speed_threshold)
        VALUES ('Snowshoes', True, 0.1)
        """
    )


def downgrade():
    op.execute(
        """
        DELETE FROM sports
        WHERE label = 'Snowshoes';
        """
    )



from typing import Dict, Optional, Tuple

from flask import current_app

from fittrackee.comments.models import Comment
from fittrackee.dates import get_date_string_for_user
from fittrackee.emails.tasks import send_email
from fittrackee.users.models import User
from fittrackee.users.utils.language import get_language
from fittrackee.workouts.models import Workout

from .exceptions import InvalidReportActionException
from .models import Report, ReportAction


class ReportEmailService:
    @staticmethod
    def _get_email_data(
        report: Report,
        reason: Optional[str] = None,
        with_user_image: bool = False,
    ) -> Tuple[Dict, Dict, str]:
        fittrackee_url = current_app.config["UI_URL"]
        reported_user: User = report.reported_user
        user_data = {
            "language": get_language(reported_user.language),
            "email": reported_user.email,
        }
        email_data = {
            "username": reported_user.username,
            "fittrackee_url": fittrackee_url,
            "reason": reason,
        }
        if with_user_image:
            email_data["user_image_url"] = (
                f"{fittrackee_url}/api/users/{reported_user.username}/picture"
                if reported_user.picture
                else f"{fittrackee_url}/img/user.png"
            )
        return user_data, email_data, fittrackee_url

    @staticmethod
    def _get_comment_email_data(
        email_data: Dict,
        comment: Comment,
        reported_user: User,
        fittrackee_url: str,
    ) -> Dict:
        comment_email_data = {
            **email_data,
            "created_at": get_date_string_for_user(
                comment.created_at, reported_user
            ),
            "text": comment.text,
        }
        if comment.workout_id:
            comment_email_data["comment_url"] = (
                f"{fittrackee_url}/workouts/{comment.workout.short_id}"
                f"/comments/{comment.short_id}"
            )
        else:
            comment_email_data["comment_url"] = (
                f"{fittrackee_url}/comments/{comment.short_id}"
            )

        return comment_email_data

    @staticmethod
    def _get_workout_email_data(
        email_data: Dict,
        workout: Workout,
        reported_user: User,
        fittrackee_url: str,
    ) -> Dict:
        return {
            **email_data,
            "map": (
                f"{fittrackee_url}/api/workouts/map/{workout.map_id}"
                if workout.map_id
                else None
            ),
            "title": workout.title,
            "workout_date": get_date_string_for_user(
                workout.workout_date, reported_user
            ),
            "workout_url": f"{fittrackee_url}/workouts/{workout.short_id}",
        }

    def _send_user_suspension_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason
        )
        email_data["appeal_url"] = f"{fittrackee_url}/profile/suspension"
        send_email.send(user_data, email_data, template="user_suspension")

    def _send_user_unsuspension_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, _ = self._get_email_data(report, reason)
        email_data["without_user_action"] = True
        send_email.send(user_data, email_data, template="user_unsuspension")

    def _send_user_warning_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )
        if not report_action:
            raise InvalidReportActionException("invalid action action")

        if report.reported_comment_id:
            email_data = self._get_comment_email_data(
                email_data,
                report.reported_comment,
                report.reported_user,
                fittrackee_url,
            )
        elif report.reported_workout_id:
            email_data = self._get_workout_email_data(
                email_data,
                report.reported_workout,
                report.reported_user,
                fittrackee_url,
            )
        email_data["appeal_url"] = (
            f"{fittrackee_url}/profile/moderation/"
            f"sanctions/{report_action.short_id}"
        )
        send_email.send(user_data, email_data, template="user_warning")

    def _send_user_warning_lifting_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )
        if not report_action:
            raise InvalidReportActionException("invalid action action")

        if report.reported_comment_id:
            email_data = self._get_comment_email_data(
                email_data,
                report.reported_comment,
                report.reported_user,
                fittrackee_url,
            )
        elif report.reported_workout_id:
            email_data = self._get_workout_email_data(
                email_data,
                report.reported_workout,
                report.reported_user,
                fittrackee_url,
            )
        email_data["without_user_action"] = True
        send_email.send(user_data, email_data, template="user_warning_lifting")

    def _send_comment_suspension_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )
        email_data = self._get_comment_email_data(
            email_data,
            report.reported_comment,
            report.reported_user,
            fittrackee_url,
        )
        send_email.send(user_data, email_data, template="comment_suspension")

    def _send_comment_unsuspension_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )
        email_data = self._get_comment_email_data(
            email_data,
            report.reported_comment,
            report.reported_user,
            fittrackee_url,
        )
        email_data["without_user_action"] = True
        send_email.send(user_data, email_data, template="comment_unsuspension")

    def _send_workout_suspension_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )
        email_data = self._get_workout_email_data(
            email_data,
            report.reported_workout,
            report.reported_user,
            fittrackee_url,
        )
        send_email.send(user_data, email_data, template="workout_suspension")

    def _send_workout_unsuspension_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )
        email_data = self._get_workout_email_data(
            email_data,
            report.reported_workout,
            report.reported_user,
            fittrackee_url,
        )
        email_data["without_user_action"] = True
        send_email.send(user_data, email_data, template="workout_unsuspension")

    def _send_appeal_rejected_email(
        self,
        *,
        report: Report,
        reason: Optional[str],
        report_action: Optional[ReportAction],
    ) -> None:
        user_data, email_data, fittrackee_url = self._get_email_data(
            report, reason, with_user_image=True
        )

        if not report_action:
            raise InvalidReportActionException("invalid action action")

        if report.reported_comment_id:
            email_data = self._get_comment_email_data(
                email_data,
                report.reported_comment,
                report.reported_user,
                fittrackee_url,
            )
        elif report.reported_workout_id:
            email_data = self._get_workout_email_data(
                email_data,
                report.reported_workout,
                report.reported_user,
                fittrackee_url,
            )
        email_data["without_user_action"] = True
        email_data["action_type"] = report_action.action_type
        send_email.send(user_data, email_data, template="appeal_rejected")

    def send_report_action_email(
        self,
        report: Report,
        action_type: str,
        reason: Optional[str],
        report_action: Optional[
            ReportAction
        ] = None,  # needed only for user warning and appeal
    ) -> None:
        send_email_func = getattr(self, f"_send_{action_type}_email")
        send_email_func(
            report=report, reason=reason, report_action=report_action
        )



import os
import secrets
import zipfile
from abc import abstractmethod
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from io import BytesIO
from logging import getLogger
from typing import IO, TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from flask import current_app

from fittrackee import db
from fittrackee.equipments.exceptions import InvalidEquipmentsException
from fittrackee.equipments.models import Equipment
from fittrackee.files import get_absolute_file_path
from fittrackee.users.models import Notification, User, UserTask
from fittrackee.visibility_levels import get_calculated_visibility
from fittrackee.workouts.models import (
    DESCRIPTION_MAX_CHARACTERS,
    NOTES_MAX_CHARACTERS,
)

from ..exceptions import (
    WorkoutExceedingValueException,
    WorkoutException,
    WorkoutFileException,
    WorkoutNoFileException,
)
from .base_workout_service import (
    BaseWorkoutService,
)
from .mixins import WorkoutFileMixin
from .workout_from_file.services import WORKOUT_FROM_FILE_SERVICES

if TYPE_CHECKING:
    from werkzeug.datastructures import FileStorage

    from fittrackee.visibility_levels import VisibilityLevel
    from fittrackee.workouts.models import Workout

appLog = getLogger("fittrackee_workouts_upload")


@dataclass
class WorkoutsData:
    sport_id: int
    analysis_visibility: Optional["VisibilityLevel"] = None
    description: Optional[str] = None
    equipment_ids: Optional[List[str]] = None
    map_visibility: Optional["VisibilityLevel"] = None
    notes: Optional[str] = None
    title: Optional[str] = None
    workout_visibility: Optional["VisibilityLevel"] = None


class AbstractWorkoutsCreationService(BaseWorkoutService, WorkoutFileMixin):
    def __init__(
        self,
        auth_user: "User",
        workouts_data: Dict,
        file: Optional["FileStorage"] = None,
    ):
        super().__init__(
            auth_user,
            workouts_data.get("sport_id"),
            workouts_data.get("equipment_ids"),
        )
        self.file: Optional["FileStorage"] = file
        self.workouts_data = WorkoutsData(**workouts_data)

    @abstractmethod
    def process(self) -> Tuple[List["Workout"], Dict]:
        pass

    def _get_user_path(
        self,
    ) -> str:
        user_path = os.path.join("workouts", str(self.auth_user.id))
        absolute_user_path = get_absolute_file_path(user_path)
        if not os.path.exists(absolute_user_path):
            os.makedirs(absolute_user_path)
        return user_path

    def get_file_path(
        self,
        workout_date: str,
        extension: str,
    ) -> str:
        suffix = secrets.token_urlsafe(8)
        new_filename = (
            f"{workout_date}_{self.workouts_data.sport_id}_{suffix}{extension}"
        )
        dir_path = self._get_user_path()
        file_path = os.path.join(dir_path, new_filename.split("/")[-1])
        return file_path

    def _get_workout_title(
        self, service_workout_name: Optional[str], workout_date: datetime
    ) -> Optional[str]:
        title = ""
        if self.workouts_data.title:
            title = self.workouts_data.title
        elif service_workout_name:
            title = service_workout_name
        return self._get_title(workout_date, title)

    def _get_workout_description(
        self, service_workout_description: Optional[str]
    ) -> Optional[str]:
        if self.workouts_data.description:
            return self.workouts_data.description[:DESCRIPTION_MAX_CHARACTERS]

        if service_workout_description:
            return service_workout_description[:DESCRIPTION_MAX_CHARACTERS]

        return None

    def _get_workout_notes(self) -> Optional[str]:
        if self.workouts_data.notes is None:
            return None
        return self.workouts_data.notes[:NOTES_MAX_CHARACTERS]

    def _store_file(
        self,
        new_workout: "Workout",
        workout_file: Optional["IO[bytes]"],
        *,
        gpx_xml_content: Optional[str],  # for non-gpx files
        extension: str = ".gpx",
        relative_path: Optional[str] = None,
    ) -> str:
        if not workout_file and not self.file:
            return ""

        if not relative_path:
            relative_path = self.get_file_path(
                workout_date=new_workout.workout_date.strftime(
                    "%Y-%m-%d_%H-%M-%S"
                ),
                extension=extension,
            )

        if extension == ".gpx":
            new_workout.gpx = relative_path

        absolute_workout_filepath = get_absolute_file_path(relative_path)
        try:
            if gpx_xml_content:
                with open(absolute_workout_filepath, "w") as f:
                    f.write(gpx_xml_content)
            elif workout_file:  # when file from archive
                with open(absolute_workout_filepath, "wb") as f:
                    workout_file.seek(0)
                    f.write(workout_file.read())
            elif self.file:
                self.file.stream.seek(0)
                self.file.save(absolute_workout_filepath)
            else:
                return ""
        except Exception as e:
            error = "error when storing gpx file"
            appLog.exception(error)
            raise WorkoutException("error", error) from e
        return absolute_workout_filepath

    def _get_archive_content(self) -> Union[BytesIO, IO[bytes]]:
        if not self.file:
            raise WorkoutNoFileException()

        # handle Python < 3.11, see:
        # - https://github.com/python/cpython/issues/70363
        # - https://docs.python.org/3.11/whatsnew/3.11.html#tempfile
        if not hasattr(self.file.stream, "seekable"):
            return BytesIO(self.file.getvalue())
        return self.file.stream

    def create_workout_from_file(
        self,
        extension: str,
        equipments: Union[List["Equipment"], None],
        workout_file: Optional["IO[bytes]"] = None,
        get_weather: bool = True,
    ) -> "Workout":
        """
        Return map absolute file path in order to delete file on error
        """
        if workout_file is None and self.file is None:
            raise WorkoutNoFileException()

        if extension == "kmz" and workout_file is None:
            workout_file = self._get_archive_content()

        workout_service = WORKOUT_FROM_FILE_SERVICES[extension](
            auth_user=self.auth_user,
            workout_file=(
                self.file.stream  # type: ignore
                if workout_file is None
                else workout_file
            ),
            sport_id=self.workouts_data.sport_id,
            stopped_speed_threshold=self.stopped_speed_threshold,
            get_weather=get_weather,
            workout=None,
        )

        # extract and calculate data from provided file
        try:
            new_workout = workout_service.process_workout()
        except (WorkoutExceedingValueException, WorkoutFileException) as e:
            appLog.exception(f"workout exception: {e!s}")
            db.session.rollback()
            raise e
        except Exception as e:
            appLog.exception(f"exception: {e!s}")
            db.session.rollback()
            if (
                "duplicate key value violates unique constraint "
                '"workout_id_start_date_unique"' in str(e)
            ):
                raise WorkoutFileException(
                    "error", "some segments have same start date"
                ) from e
            raise WorkoutException(
                "error", "error when processing workout"
            ) from e

        # store title, description and notes
        new_workout.title = self._get_workout_title(
            workout_service.workout_name, new_workout.workout_date
        )
        new_workout.description = self._get_workout_description(
            workout_service.workout_description
        )
        new_workout.notes = self._get_workout_notes()

        # add equipments if ids provided
        if equipments is not None:
            new_workout.equipments = equipments

        # update visibility
        new_workout.workout_visibility = (
            self.workouts_data.workout_visibility
            if self.workouts_data.workout_visibility
            else self.auth_user.workouts_visibility
        )
        new_workout.analysis_visibility = get_calculated_visibility(
            visibility=(
                self.workouts_data.analysis_visibility
                if self.workouts_data.analysis_visibility
                else self.auth_user.analysis_visibility
            ),
            parent_visibility=new_workout.workout_visibility,
        )
        new_workout.map_visibility = get_calculated_visibility(
            visibility=(
                self.workouts_data.map_visibility
                if self.workouts_data.map_visibility
                else self.auth_user.map_visibility
            ),
            parent_visibility=new_workout.analysis_visibility,
        )

        # store workout file
        absolute_workout_filepath = self._store_file(
            new_workout,
            workout_file,
            gpx_xml_content=(
                None if extension == "gpx" else workout_service.gpx.to_xml()
            ),
        )

        if extension == "gpx":
            new_workout.original_file = new_workout.gpx
        # store original file if extension is not .gpx
        elif new_workout.gpx:
            original_extension = f".{extension}"
            relative_path = new_workout.gpx.replace(".gpx", original_extension)
            self._store_file(
                new_workout,
                workout_file,
                gpx_xml_content=None,
                extension=original_extension,
                relative_path=relative_path,
            )
            new_workout.original_file = relative_path

        # generate and store map image
        map_filepath = self.get_file_path(
            workout_date=new_workout.workout_date.strftime(
                "%Y-%m-%d_%H-%M-%S"
            ),
            extension=".png",
        )
        new_workout.map = map_filepath
        absolute_map_filepath = get_absolute_file_path(map_filepath)
        try:
            workout_service.generate_map_image(
                map_filepath=absolute_map_filepath,
                coordinates=workout_service.coordinates,
            )
            new_workout.map_id = workout_service.get_map_hash(map_filepath)
        except Exception as e:
            if os.path.exists(absolute_map_filepath):
                os.remove(absolute_map_filepath)
            if os.path.exists(absolute_workout_filepath):
                os.remove(absolute_workout_filepath)
            raise WorkoutException(
                "error", "error when generating map image"
            ) from e
        db.session.commit()
        return new_workout

    def process_archive_content(
        self,
        archive_content: Union[BytesIO, IO[bytes]],
        files_to_process: List[str],
        equipments: Union[List["Equipment"], None],
        upload_task: Optional["UserTask"] = None,
        get_weather: bool = True,
    ) -> Tuple[List["Workout"], Dict]:
        if not files_to_process:
            raise WorkoutFileException(
                "error", "No files from archive to process"
            )
        appLog.debug(" > starting archive processing...")

        new_workouts = []
        errored_workouts = {}
        total_files = len(files_to_process)
        with zipfile.ZipFile(archive_content, "r") as zip_ref:
            for index, file in enumerate(files_to_process, start=1):
                appLog.debug(f"  - file {index}/{total_files}")
                try:
                    extension = self._get_extension(file)
                    file_content = zip_ref.open(file)
                    new_workout = self.create_workout_from_file(
                        extension, equipments, file_content, get_weather
                    )
                except Exception as e:
                    db.session.rollback()
                    error = e.args[0]
                    errored_workouts[file] = error
                    appLog.debug(f"    > error occurred: {error}")
                    if upload_task:
                        upload_task.progress = int(100 * index / total_files)
                        db.session.commit()
                    continue

                new_workouts.append(new_workout)
                appLog.debug("    > upload done")
                if upload_task:
                    upload_task.data = {
                        **upload_task.data,
                        "new_workouts_count": len(new_workouts),
                    }
                    upload_task.progress = int(100 * index / total_files)
                    db.session.commit()

        return new_workouts, errored_workouts


class WorkoutsFromFileCreationService(AbstractWorkoutsCreationService):
    def __init__(
        self,
        auth_user: "User",
        workouts_data: Dict,
        file: Optional["FileStorage"] = None,
    ):
        super().__init__(auth_user, workouts_data, file)

    def get_files_from_archive(self) -> List[str]:
        if not self.file:
            raise WorkoutNoFileException()
        files_to_process = []
        max_file_size = current_app.config["max_single_file_size"]
        try:
            with zipfile.ZipFile(self.file.stream, "r") as zip_ref:
                for zip_info in zip_ref.infolist():
                    if self._is_workout_file(zip_info.filename):
                        if zip_info.file_size > max_file_size:
                            raise WorkoutFileException(
                                "fail",
                                (
                                    "at least one file in zip archive "
                                    "exceeds size limit, "
                                    "please check the archive"
                                ),
                            )
                        files_to_process.append(zip_info.filename)

                        if (
                            len(files_to_process)
                            > current_app.config["file_limit_import"]
                        ):
                            raise WorkoutFileException(
                                "fail",
                                (
                                    "the number of files in the archive "
                                    "exceeds the limit"
                                ),
                            )

        except zipfile.BadZipFile as e:
            raise WorkoutFileException("error", "invalid zip file") from e

        if not files_to_process:
            raise WorkoutFileException(
                "fail", "archive does not contain valid workout files"
            )

        return files_to_process

    def add_workouts_upload_task(
        self,
        files_to_process: List[str],
        equipments: Union[List["Equipment"], None],
    ) -> str:
        from fittrackee.workouts.tasks import upload_workouts_archive

        if self.file is None:
            raise WorkoutNoFileException()
        if not files_to_process:
            raise WorkoutFileException(
                "error", "No files from archive to process"
            )
        if UserTask.query.filter(
            UserTask.user_id == self.auth_user.id,
            UserTask.task_type == "workouts_archive_upload",
            UserTask.aborted == False,  # noqa
            UserTask.errored == False,  # noqa
            UserTask.progress != 100,
        ).first():
            raise WorkoutException("invalid", "ongoing upload task exists")

        relative_path = os.path.join(
            self._get_user_path(),
            f"archive_{secrets.token_urlsafe(20)}.zip",
        )
        path = get_absolute_file_path(relative_path)
        self.file.stream.seek(0)
        self.file.save(path)

        upload_task = UserTask(
            user_id=self.auth_user.id,
            task_type="workouts_archive_upload",
            data={
                "workouts_data": asdict(self.workouts_data),
                "files_to_process": files_to_process,
                "new_workouts_count": 0,
                "equipment_ids": (
                    None
                    if equipments is None
                    else [equipment.id for equipment in equipments]
                ),
                "original_file_name": self.file.filename,
            },
            file_path=relative_path,
        )
        upload_task.errors = {
            "archive": None,
            "files": {},
        }
        upload_task.file_size = os.path.getsize(path)
        db.session.add(upload_task)
        db.session.commit()

        message = upload_workouts_archive.send(task_id=upload_task.id)
        upload_task.message_id = message.message_id
        db.session.commit()

        return upload_task.short_id

    def process_zip_archive(
        self, equipments: Union[List["Equipment"], None]
    ) -> Tuple[List["Workout"], Dict]:
        if not self.file:
            raise WorkoutNoFileException()
        files_to_process = self.get_files_from_archive()
        if (
            len(files_to_process)
            > current_app.config["file_sync_limit_import"]
        ):
            task_short_id = self.add_workouts_upload_task(
                files_to_process, equipments
            )
            return [], {"task_short_id": task_short_id}

        new_workouts, errored_workouts = self.process_archive_content(
            archive_content=self._get_archive_content(),
            files_to_process=files_to_process,
            equipments=equipments,
        )
        return new_workouts, {
            "errored_workouts": errored_workouts,
            "task_short_id": None,
        }

    def process(self) -> Tuple[List["Workout"], Dict]:
        if self.file is None:
            raise WorkoutNoFileException()
        if self.file.filename is None:
            raise WorkoutFileException("error", "workout file has no filename")

        extension = self._get_extension(self.file.filename)
        if extension not in current_app.config["WORKOUT_ALLOWED_EXTENSIONS"]:
            raise WorkoutFileException(
                "error", "workout file invalid extension"
            )
        equipments = self.get_equipments()

        try:
            if extension == "zip":
                new_workouts, processing_output = self.process_zip_archive(
                    equipments
                )
                return new_workouts, processing_output
            else:
                new_workout = self.create_workout_from_file(
                    extension, equipments
                )
                return [new_workout], {
                    "task_short_id": None,
                    "errored_workouts": [],
                }
        except InvalidEquipmentsException as e:
            raise WorkoutException("error", str(e)) from e


class WorkoutsFromArchiveCreationAsyncService(AbstractWorkoutsCreationService):
    def __init__(
        self,
        task_id: int,
    ) -> None:
        upload_task = UserTask.query.filter_by(id=task_id).first()
        if (
            not upload_task
            or upload_task.task_type != "workouts_archive_upload"
        ):
            raise WorkoutException("error", "no import task found")
        auth_user = User.query.filter_by(id=upload_task.user_id).one()
        import_data = upload_task.data
        super().__init__(auth_user, import_data["workouts_data"], None)
        self.file_path = get_absolute_file_path(upload_task.file_path)
        self.files_to_process = import_data["files_to_process"]
        self.equipment_ids = import_data["equipment_ids"]
        self.upload_task = upload_task

    def process(self) -> Tuple[List["Workout"], Dict]:
        if self.equipment_ids is None:
            equipments = None
        elif not self.equipment_ids:
            equipments = []
        else:
            equipments = Equipment.query.filter(
                Equipment.id.in_(self.equipment_ids)
            ).all()

        errored_workouts: Dict = {}
        new_workouts: List["Workout"] = []
        try:
            with open(self.file_path, "rb") as zip_file:
                new_workouts, errored_workouts = self.process_archive_content(
                    archive_content=zip_file,
                    files_to_process=self.files_to_process,
                    equipments=equipments,
                    upload_task=self.upload_task,
                    get_weather=False,
                )
        except FileNotFoundError:
            self.upload_task.errored = True
            self.upload_task.errors = {
                **self.upload_task.errors,
                "archive": "archive file does not exist",
            }
        except Exception:
            self.upload_task.errored = True
            self.upload_task.errors = {
                **self.upload_task.errors,
                "archive": "error during archive processing",
            }

        if errored_workouts:
            self.upload_task.errored = True
            self.upload_task.errors = {
                **self.upload_task.errors,
                "files": errored_workouts,
            }
        db.session.commit()

        if os.path.exists(self.file_path):
            os.remove(self.file_path)

        notification = Notification(
            from_user_id=self.upload_task.user_id,
            to_user_id=self.upload_task.user_id,
            created_at=datetime.now(tz=timezone.utc),
            event_object_id=self.upload_task.id,
            event_type=self.upload_task.task_type,
        )
        db.session.add(notification)
        db.session.commit()

        return new_workouts, {
            "errored_workouts": errored_workouts,
            "task_short_id": self.upload_task.short_id,
        }



import logging
import sys
from datetime import datetime
from typing import Optional

import click

from fittrackee.cli.app import app
from fittrackee.users.models import User
from fittrackee.workouts.constants import WORKOUT_ALLOWED_EXTENSIONS
from fittrackee.workouts.models import Sport
from fittrackee.workouts.services.workouts_from_file_refresh_service import (
    WorkoutsFromFileRefreshService,
)
from fittrackee.workouts.tasks import (
    process_workouts_archive_upload,
    process_workouts_archives_uploads,
)
from fittrackee.workouts.utils.workouts import get_workout_datetime

WORKOUT_VALID_EXTENSIONS = ", ".join(WORKOUT_ALLOWED_EXTENSIONS)
handler = logging.StreamHandler()
logger = logging.getLogger("fittrackee_workouts_cli")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


def validate_order(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    if value and value not in ["asc", "desc"]:
        raise click.BadParameter("order must be 'asc' or 'desc'")
    return value


def validate_extension(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    if value and value not in WORKOUT_ALLOWED_EXTENSIONS:
        raise click.BadParameter(
            f"valid extensions are: {WORKOUT_VALID_EXTENSIONS}"
        )
    return value


def validate_sport_id(
    ctx: click.core.Context, param: click.core.Option, value: Optional[int]
) -> Optional[int]:
    with app.app_context():
        if value is not None and not Sport.query.filter_by(id=value).first():
            raise click.BadParameter(f"invalid sport id '{value}'")
    return value


def validate_user(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional[str]:
    with app.app_context():
        if (
            value is not None
            and not User.query.filter_by(username=value).first()
        ):
            raise click.BadParameter(f"user '{value}' does not exist")
    return value


def validate_number(
    ctx: click.core.Context, param: click.core.Option, value: Optional[int]
) -> Optional[int]:
    if value is not None and value < 1:
        raise click.BadParameter("value must be greater than 0")

    return value


def validate_date(
    ctx: click.core.Context, param: click.core.Option, value: Optional[str]
) -> Optional["datetime"]:
    if value is not None:
        try:
            date_value, _ = get_workout_datetime(
                workout_date=value,
                user_timezone=None,
                date_str_format="%Y-%m-%d",
            )
            return date_value
        except Exception as e:
            raise click.BadParameter(
                f"'{value}' does not match format '%Y-%m-%d'"
            ) from e
    return value


@click.group(name="workouts")
def workouts_cli() -> None:
    """Manage workouts."""
    pass


@workouts_cli.command("archive_uploads")
@click.option(
    "--max",
    "max_archives",
    type=int,
    required=True,
    help="Maximum number of workouts archive to process.",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled).",
)
def workouts_archive_uploads(max_archives: int, verbose: bool) -> None:
    """
    Process workouts archive uploads if queued tasks exist (progress = 0 and
    not aborted/errored).
    To use in case redis is not set.
    """
    with app.app_context():
        if verbose:
            upload_logger = logging.getLogger("fittrackee_workouts_upload")
            upload_logger.setLevel(logging.DEBUG)
            upload_logger.addHandler(handler)
        count = process_workouts_archives_uploads(max_archives, logger)
        logger.info(f"\nWorkouts archives processed: {count}.")


@workouts_cli.command("archive_upload")
@click.option(
    "--id",
    "task_short_id",
    type=str,
    required=True,
    help="Id of task to process",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled).",
)
def process_queued_archive_upload(task_short_id: str, verbose: bool) -> None:
    """
    Process given queued workouts archive upload.

    To use in case redis is not set.
    """
    with app.app_context():
        if verbose:
            upload_logger = logging.getLogger("fittrackee_workouts_upload")
            upload_logger.setLevel(logging.DEBUG)
            upload_logger.addHandler(handler)
        try:
            process_workouts_archive_upload(task_short_id, logger)
        except Exception as e:
            logger.error(str(e))
            sys.exit(1)
        logger.info("\nDone.")


@workouts_cli.command("refresh")
@click.option(
    "--sport-id",
    help="sport id",
    type=int,
    callback=validate_sport_id,
)
@click.option(
    "--from",
    "date_from",
    help="start date (format: %Y-%m-%d)",
    callback=validate_date,
)
@click.option(
    "--to",
    "date_to",
    help="end date (format: %Y-%m-%d)",
    callback=validate_date,
)
@click.option(
    "--per-page",
    help="number of workouts per page (default: 10)",
    type=int,
    callback=validate_number,
    default=10,
)
@click.option(
    "--page",
    help="page number (default: 1)",
    type=int,
    callback=validate_number,
    default=1,
)
@click.option(
    "--order",
    help="workout date order: 'asc' or 'desc' (default: 'asc')",
    type=str,
    default="asc",
    callback=validate_order,
)
@click.option(
    "--user",
    help="username of workouts owner",
    type=str,
    callback=validate_user,
)
@click.option(
    "--extension",
    help=(
        "workout file extension "
        f"(valid values are: {WORKOUT_VALID_EXTENSIONS})"
    ),
    type=str,
    callback=validate_extension,
)
@click.option(
    "--with-weather",
    help=(
        "enable weather data collection if weather provider is set and "
        "workout has no weather data. "
        "WARNING: depending on subscription, the rate limit can be reached, "
        "leading to errors and preventing weather data being collected during "
        "next uploads until the limit is reset (default: disabled)"
    ),
    is_flag=True,
    show_default=True,
    default=False,
)
@click.option(
    "--add-missing-geometry",
    help=(
        "if provided, it refreshes only workouts without geometry in database "
        "to add geometry and points. This option is provided to update "
        "workouts created before v1.x and will be removed in a future version "
        "when all workouts must have geometry."
    ),
    is_flag=True,
    show_default=True,
    default=False,
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled)",
)
def refresh_workouts(
    sport_id: Optional[int] = None,
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
    per_page: int = 10,
    page: int = 1,
    order: str = "asc",
    user: Optional[str] = None,
    extension: Optional[str] = None,
    with_weather: bool = False,
    add_missing_geometry: bool = False,
    verbose: bool = False,
) -> None:
    """
    Refresh workouts
    """
    with app.app_context():
        if verbose:
            refresh_logger = logging.getLogger("fittrackee_workouts_refresh")
            refresh_logger.setLevel(logging.DEBUG)
            refresh_logger.addHandler(handler)

        try:
            service = WorkoutsFromFileRefreshService(
                sport_id=sport_id,
                date_from=date_from,
                date_to=date_to,
                per_page=per_page,
                page=page,
                order=order,
                user=user,
                extension=extension,
                with_weather=with_weather,
                verbose=verbose,
                add_geometry=add_missing_geometry,
                logger=logger,
            )
            service.refresh()
        except Exception as e:
            logger.error(str(e))
            sys.exit(1)

        logger.info("\nDone.")



class InvalidReportActionAppealException(Exception):
    pass


class InvalidReportActionAppealUserException(Exception):
    pass


class InvalidReportActionException(Exception):
    pass


class InvalidReportException(Exception):
    pass


class InvalidReporterException(Exception):
    pass


class ReportActionAppealForbiddenException(Exception):
    pass


class ReportActionForbiddenException(Exception):
    pass


class ReportCommentForbiddenException(Exception):
    pass


class ReportForbiddenException(Exception):
    pass


class ReportNotFoundException(Exception):
    pass


class SuspendedObjectException(Exception):
    def __init__(self, object_type: str) -> None:
        super().__init__(f"{object_type} already suspended")
        self.object_type = object_type


class UserWarningExistsException(Exception):
    pass



from typing import IO

import gpxpy.gpx
import xmltodict
from gpxpy.gpxfield import parse_time

from ...exceptions import WorkoutFileException
from .workout_gpx_service import WorkoutGpxService


class WorkoutKmlService(WorkoutGpxService):
    @classmethod
    def parse_file(
        cls,
        workout_file: IO[bytes],
        segments_creation_event: str,
    ) -> "gpxpy.gpx.GPX":
        """
        Only kml files with Placemark/MultiTrack/Tracks are supported.
        Files with folders or multiple Placemark are no supported.

        For now, a gpx file generated from kml file contains one track (<trk>)
        corresponding to the first MultiTrack, containing one segment
        (<trkseg>) per kml track (<Track>).

        Tested with files generated with OpenTracks.

        Note:
        - segments_creation_event is not used (only for .fit files)
        """
        try:
            kml_dict = xmltodict.parse(workout_file)
        except Exception as e:
            raise WorkoutFileException(
                "error", "error when parsing kml file"
            ) from e

        placemarks = kml_dict["kml"].get("Document", {}).get("Placemark", [])
        if not placemarks:
            raise WorkoutFileException(
                "error", "unsupported kml file"
            ) from None
        if isinstance(placemarks, dict):
            placemarks = [placemarks]
        placemark = placemarks[0]

        kml_tracks = placemark.get("MultiTrack", {}).get("Track", [])
        coordinates_key = "coord"
        if not kml_tracks:
            kml_tracks = placemark.get("gx:MultiTrack", {}).get("gx:Track", [])
            if not kml_tracks:
                raise WorkoutFileException(
                    "error", "no tracks in kml file"
                ) from None
            coordinates_key = "gx:coord"

        if isinstance(kml_tracks, dict):
            kml_tracks = [kml_tracks]

        gpx_track = gpxpy.gpx.GPXTrack()
        gpx_track.name = placemark.get("name")
        gpx_track.description = placemark.get("description")
        for kml_track in kml_tracks:
            gpx_segment = gpxpy.gpx.GPXTrackSegment()

            coords = kml_track.get(coordinates_key, [])
            if not coords:
                continue
            if isinstance(coords, dict):
                coords = [coords]
            times = kml_track.get("when", [])
            if isinstance(times, dict):
                times = [times]

            heart_rates = []
            cadences = []
            powers = []
            extended_data = (
                kml_track.get("ExtendedData", {})
                .get("SchemaData", {})
                .get("gx:SimpleArrayData", [])
            )
            for data in extended_data:
                if data["@name"] == "heartrate":
                    heart_rates = data["gx:value"]
                if data["@name"] == "cadence":
                    cadences = data["gx:value"]
                if data["@name"] == "power":
                    powers = data["gx:value"]

            for index, date in enumerate(times):
                if not coords[index]:
                    continue
                longitude, latitude, elevation = coords[index].split()
                point = gpxpy.gpx.GPXTrackPoint(
                    longitude=float(longitude),
                    latitude=float(latitude),
                    elevation=float(elevation),
                    time=parse_time(date),
                )
                heart_rate = (
                    heart_rates[index] if len(heart_rates) > index else None
                )
                cadence = cadences[index] if len(cadences) > index else None
                power = powers[index] if len(powers) > index else None
                if (
                    heart_rate is not None
                    or cadence is not None
                    or power is not None
                ):
                    point.extensions.append(
                        cls._get_extensions(heart_rate, cadence, power)
                    )

                gpx_segment.points.append(point)
            gpx_track.segments.append(gpx_segment)
        gpx = gpxpy.gpx.GPX()
        gpx.tracks.append(gpx_track)
        return gpx



"""add uuid to activities

Revision ID: 3243cd25eca7
Revises: 8a0aad4c838c
Create Date: 2020-12-30 14:54:45.568864

"""
from uuid import uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '3243cd25eca7'
down_revision = '8a0aad4c838c'
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    activities_helper = sa.Table(
        'activities',
        sa.MetaData(),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False),
    )

    op.add_column(
        'activities',
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=True),
    )
    op.create_unique_constraint('activities_uuid_key', 'activities', ['uuid'])
    op.add_column(
        'activity_segments',
        sa.Column(
            'activity_uuid', postgresql.UUID(as_uuid=True), nullable=True
        ),
    )
    op.add_column(
        'records',
        sa.Column(
            'activity_uuid', postgresql.UUID(as_uuid=True), nullable=True
        ),
    )
    for activity in connection.execute(activities_helper.select()):
        activity_uuid = uuid4()
        op.execute(
            f"UPDATE activities SET uuid = '{activity_uuid}' "
            f"WHERE activities.id = {activity.id}"
        )
        op.execute(
            f"UPDATE records SET activity_uuid = '{activity_uuid}' "
            f"WHERE records.activity_id = {activity.id}"
        )
        op.execute(
            f"UPDATE activity_segments SET activity_uuid = '{activity_uuid}' "
            f"WHERE activity_segments.activity_id = {activity.id}"
        )

    op.alter_column('activities', 'uuid', nullable=False)
    op.alter_column('activity_segments', 'activity_uuid', nullable=False)
    op.alter_column('records', 'activity_uuid', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('records', 'activity_uuid')
    op.drop_column('activity_segments', 'activity_uuid')
    op.drop_constraint('activities_uuid_key', 'activities', type_='unique')
    op.drop_column('activities', 'uuid')
    # ### end Alembic commands ###



import re
from typing import IO, Dict, Optional

import gpxpy.gpx
import xmltodict
from gpxpy.gpxfield import parse_time

from ...exceptions import WorkoutFileException
from .constants import NSMAP
from .workout_gpx_service import WorkoutGpxService


class WorkoutTcxService(WorkoutGpxService):
    @staticmethod
    def _get_elevation(point: Dict) -> Optional[float]:
        altitude_meters = point.get("AltitudeMeters")
        if altitude_meters is None:
            return None
        elevation = float(altitude_meters)
        # workaround
        # some devices/softwares return invalid elevation values
        # see https://github.com/piggz/harbour-amazfish/issues/494
        if -9999.99 < elevation < 9999.99:
            return elevation
        return None

    @classmethod
    def parse_file(
        cls, workout_file: IO[bytes], segments_creation_event: str
    ) -> "gpxpy.gpx.GPX":
        """
        Tcx files contain activities that contain laps containing tracks.
        A gpx file generated from tcx file contains one track containing one
        segment per activity.

        TODO:
        - handle multiple sports activities like Swimrun

        Note:
        - segments_creation_event is not used (only for .fit files)
        """
        try:
            tcx_dict = xmltodict.parse(workout_file)
        except Exception as e:
            raise WorkoutFileException(
                "error", "error when parsing tcx file"
            ) from e

        activities = (
            tcx_dict["TrainingCenterDatabase"]
            .get("Activities", {})
            .get("Activity", [])
        )
        if isinstance(activities, dict):
            activities = [activities]

        if not activities:
            raise WorkoutFileException(
                "error", "no activities in tcx file"
            ) from None

        gpx_track = gpxpy.gpx.GPXTrack()
        has_tracks = False
        tag_prefix = None
        creator = ""

        for activity in activities:
            has_points = False

            if not creator:
                creator = activity.get("Creator", {}).get("Name", "")

            gpx_segment = gpxpy.gpx.GPXTrackSegment()
            laps = activity.get("Lap", [])
            if isinstance(laps, dict):
                laps = [laps]
            if not laps:
                continue

            for lap in laps:
                tcx_tracks = lap.get("Track", [])
                if isinstance(tcx_tracks, dict):
                    tcx_tracks = [tcx_tracks]
                if not tcx_tracks:
                    continue

                for tcx_track in tcx_tracks:
                    has_tracks = True

                    points = tcx_track.get("Trackpoint", [])
                    if isinstance(points, dict):
                        points = [points]

                    if not points:
                        continue

                    for point in points:
                        coordinates = point.get("Position", {})
                        if not coordinates:
                            continue

                        gpx_track_point = gpxpy.gpx.GPXTrackPoint(
                            longitude=float(
                                coordinates.get("LongitudeDegrees")
                            ),
                            latitude=float(coordinates.get("LatitudeDegrees")),
                            elevation=cls._get_elevation(point),
                            time=parse_time(point.get("Time")),
                        )

                        heart_rate = point.get("HeartRateBpm", {}).get("Value")
                        cadence = point.get("Cadence")
                        power = None

                        extensions = point.get("Extensions", {})
                        tpx_tags = [
                            key
                            for key in extensions.keys()
                            if re.search("TPX$", key)
                        ]
                        if tpx_tags:
                            tpx_tag = tpx_tags[0]
                            if tag_prefix is None:
                                tag_prefix = (
                                    f"{tpx_tag.split(':')[0]}:"
                                    if ":" in tpx_tag
                                    else ""
                                )

                            power = extensions.get(tpx_tag, {}).get(
                                f"{tag_prefix}Watts"
                            )

                            if not cadence:
                                cadence = extensions.get(
                                    f"{tag_prefix}TPX", {}
                                ).get(f"{tag_prefix}RunCadence")

                        if (
                            heart_rate is not None
                            or cadence is not None
                            or power is not None
                        ):
                            gpx_track_point.extensions.append(
                                cls._get_extensions(heart_rate, cadence, power)
                            )

                        gpx_segment.points.append(gpx_track_point)
                        has_points = True

            if has_points:
                gpx_track.segments.append(gpx_segment)

        if not has_tracks:
            raise WorkoutFileException(
                "error", "no laps or no tracks in tcx file"
            ) from None

        if not gpx_track.segments:
            raise WorkoutFileException(
                "error", "no coordinates in tcx file"
            ) from None

        gpx = gpxpy.gpx.GPX()
        author = (
            tcx_dict["TrainingCenterDatabase"].get("Author", {}).get("Name")
        )
        gpx.creator = creator if creator else author
        gpx.nsmap = NSMAP
        gpx.tracks.append(gpx_track)
        return gpx



import zipfile
from datetime import datetime
from logging import Logger, getLogger
from typing import IO, TYPE_CHECKING, Optional, Union

from sqlalchemy import asc, desc, or_

from fittrackee import db
from fittrackee.files import get_absolute_file_path
from fittrackee.users.models import User, UserSportPreference
from fittrackee.workouts.models import Workout, WorkoutSegment

from ..exceptions import (
    WorkoutExceedingValueException,
    WorkoutException,
    WorkoutFileException,
    WorkoutRefreshException,
)
from .mixins import WorkoutFileMixin
from .workout_from_file.services import WORKOUT_FROM_FILE_SERVICES

if TYPE_CHECKING:
    from fittrackee.workouts.models import Sport


appLog = getLogger("fittrackee_workout_refresh")


class WorkoutFromFileRefreshService(WorkoutFileMixin):
    """
    recalculate values ang regenerate geometry and points
    """

    def __init__(self, workout: "Workout", update_weather: bool = False):
        if not workout.original_file:
            raise WorkoutRefreshException(
                "error", "workout without original file"
            )

        self.original_file: str = workout.original_file
        self.workout: "Workout" = workout
        self.user: "User" = workout.user
        self.sport: "Sport" = workout.sport
        self.sport_preferences = UserSportPreference.query.filter_by(
            user_id=self.user.id, sport_id=self.sport.id
        ).first()
        self.stopped_speed_threshold = (
            self.sport.stopped_speed_threshold
            if self.sport_preferences is None
            else self.sport_preferences.stopped_speed_threshold
        )
        self.update_weather = update_weather

    def get_file_content(self, file_extension: str) -> Union[bytes, IO[bytes]]:
        try:
            file_path = get_absolute_file_path(self.original_file)

            if file_extension == "kmz":
                with zipfile.ZipFile(file_path, "r") as kmz_ref:
                    return kmz_ref.open("doc.kml")

            with open(file_path, "rb") as f:
                file_content = f.read()
                return file_content
        except Exception:
            raise WorkoutFileException(
                "error", "error when opening original file"
            ) from None

    def refresh(self) -> "Workout":
        file_extension = self._get_extension(self.original_file)
        if not self._is_valid_workout_file_extension(file_extension):
            raise WorkoutRefreshException("error", "invalid file extension")

        file_content = self.get_file_content(file_extension)
        workout_service = WORKOUT_FROM_FILE_SERVICES[file_extension](
            auth_user=self.user,
            workout_file=file_content,  # type: ignore[arg-type]
            workout=self.workout,
            sport_id=self.sport.id,
            stopped_speed_threshold=self.stopped_speed_threshold,
            get_weather=self.update_weather,
        )

        # extract and calculate data from provided file
        # note: workout date is not updated
        try:
            workout_service.process_workout()
        except (WorkoutExceedingValueException, WorkoutFileException) as e:
            appLog.exception(f"workout exception: {e!s}")
            db.session.rollback()
            raise e
        except Exception as e:
            appLog.exception(f"exception: {e!s}")
            db.session.rollback()
            raise WorkoutException(
                "error", "error when processing workout"
            ) from e

        db.session.flush()

        # if original file is not a gpx, store the newly generated gpx file
        if file_extension != "gpx":
            gpx_file = (
                self.workout.gpx
                if self.workout.gpx
                else self.original_file.replace(file_extension, "gpx")
            )
            with open(get_absolute_file_path(gpx_file), "w") as f:
                f.write(workout_service.gpx.to_xml())

        db.session.commit()
        db.session.refresh(self.workout)
        return self.workout


class WorkoutsFromFileRefreshService:
    """
    All checks on parameters are made by the CLI command.
    """

    def __init__(
        self,
        logger: "Logger",
        per_page: Optional[int] = 10,
        page: Optional[int] = 1,
        order: Optional[str] = "asc",
        user: Optional[str] = None,
        extension: Optional[str] = None,
        sport_id: Optional[int] = None,
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        with_weather: bool = False,
        add_geometry: bool = False,
        verbose: bool = False,
    ) -> None:
        self.per_page: Optional[int] = per_page
        self.page: Optional[int] = page
        self.order: Optional[str] = order
        self.username: Optional[str] = user
        self.extension: Optional[str] = extension if extension else None
        self.sport_id: Optional[int] = sport_id
        self.date_from: Optional["datetime"] = date_from
        self.date_to: Optional["datetime"] = date_to
        self.with_weather: bool = with_weather
        self.add_geometry: bool = add_geometry
        self.logger: "Logger" = logger
        self.verbose: bool = verbose

    def _log_if_verbose(self, message: str) -> None:
        if not self.verbose:
            return
        self.logger.info(message)

    def refresh(self) -> int:
        workouts_to_refresh_query = Workout.query
        filters = [Workout.gpx != None]  # noqa
        if self.username:
            workouts_to_refresh_query = workouts_to_refresh_query.join(
                User, User.id == Workout.user_id
            )
            filters.extend([User.username == self.username])
        if self.extension:
            filters.extend([Workout.original_file.like(f"%{self.extension}")])
        if self.sport_id:
            filters.extend([Workout.sport_id == self.sport_id])
        if self.date_from:
            filters.extend([Workout.workout_date >= self.date_from])
        if self.date_to:
            filters.extend([Workout.workout_date <= self.date_to])
        if self.add_geometry:
            workouts_to_refresh_query = workouts_to_refresh_query.join(
                WorkoutSegment
            )
            filters.extend(
                [
                    or_(
                        Workout.start_point_geom == None,  # noqa
                        WorkoutSegment.geom == None,  # noqa
                    )
                ]
            )

        updated_workouts = 0
        workouts_to_refresh = (
            workouts_to_refresh_query.filter(*filters)
            .order_by(
                asc(Workout.workout_date)
                if self.order == "asc"
                else desc(Workout.workout_date)
            )
            .paginate(page=self.page, per_page=self.per_page, error_out=False)
            .items
        )

        total = len(workouts_to_refresh)
        if not total:
            self.logger.info("No workouts to refresh.")
            return 0

        self.logger.info(f"Number of workouts to refresh: {total}")
        for index, workout in enumerate(workouts_to_refresh, start=1):
            self._log_if_verbose(f"Refreshing workout {index}/{total}...")

            try:
                service = WorkoutFromFileRefreshService(
                    workout, update_weather=self.with_weather
                )
                service.refresh()
                updated_workouts += 1
            except Exception as e:
                self.logger.error(
                    (
                        f"Error when refreshing workout '{workout.short_id}' "
                        f"(user: {workout.user.username}): {e}"
                    )
                )
                continue

        if total:
            self._log_if_verbose(
                "\nRefresh done:\n"
                f"- updated workouts: {updated_workouts}\n"
                f"- errored workouts: {total - updated_workouts}"
            )

        return updated_workouts
